import asyncio
import json
from typing import Any, cast

import httpx
from pydantic import TypeAdapter

from mistralai_workflows.protocol.v1.events import (
    CustomTaskCompleted,
    CustomTaskInProgress,
    CustomTaskStarted,
    WorkflowEvent,
    WorkflowEventType,
)
from mistralai_workflows.protocol.v1.streaming import StreamEventSsePayload

SSE_DATA_PREFIX = "data:"

sse_payload_adapter: TypeAdapter[StreamEventSsePayload] = TypeAdapter(StreamEventSsePayload)
workflow_event_adapter: TypeAdapter[WorkflowEvent] = TypeAdapter(WorkflowEvent)


async def poll_workflow_status(
    client: httpx.AsyncClient,
    execution_id: str,
    expected_status: str,
    timeout_seconds: int = 30,
) -> dict[str, Any]:
    for _ in range(timeout_seconds):
        await asyncio.sleep(1)
        status_response = await client.get(f"/v1/workflows/executions/{execution_id}")
        status_response.raise_for_status()
        status_data = status_response.json()
        status = status_data.get("status")

        if status == expected_status:
            return cast(dict[str, Any], status_data)
        elif status in ["FAILED", "TERMINATED", "TIMED_OUT", "COMPLETED"]:
            if status != expected_status:
                raise RuntimeError(f"Workflow ended with status: {status}, expected: {expected_status}")

    raise TimeoutError(f"Workflow did not reach {expected_status} status within {timeout_seconds}s")


async def poll_pending_inputs(
    client: httpx.AsyncClient,
    execution_id: str,
    expected_count: int = 1,
    timeout_seconds: int = 10,
) -> list[dict[str, Any]]:
    for _ in range(timeout_seconds * 10):
        await asyncio.sleep(0.1)
        try:
            pending_response = await client.post(
                f"/v1/workflows/executions/{execution_id}/queries",
                json={"name": "__get_pending_inputs"},
            )
            pending_response.raise_for_status()
            pending_data = pending_response.json()
            pending_inputs = pending_data.get("result", {}).get("pending_inputs", [])

            if len(pending_inputs) >= expected_count:
                return cast(list[dict[str, Any]], pending_inputs)
        except (httpx.HTTPError, ValueError, KeyError):
            pass

    raise TimeoutError(f"Did not receive {expected_count} pending inputs within {timeout_seconds}s")


async def submit_workflow_update(
    client: httpx.AsyncClient,
    execution_id: str,
    update_name: str,
    update_input: dict[str, Any],
) -> None:
    update_response = await client.post(
        f"/v1/workflows/executions/{execution_id}/updates",
        json={
            "name": update_name,
            "input": update_input,
        },
    )
    update_response.raise_for_status()


async def execute_workflow(client: httpx.AsyncClient, workflow_name: str, input_data: dict[str, Any]) -> str:
    response = await client.post(
        f"/v1/workflows/{workflow_name}/execute",
        json={"input": input_data},
    )
    response.raise_for_status()
    data = response.json()
    execution_id = data.get("execution_id")
    if not execution_id:
        raise ValueError(f"No execution_id returned: {data}")
    return cast(str, execution_id)


async def stream_workflow_events(
    client: httpx.AsyncClient,
    execution_id: str,
    timeout_seconds: float = 30.0,
) -> list[dict[str, Any]]:
    received_events = []
    async with client.stream(
        "GET",
        "/v1/workflows/events/stream",
        params={"workflow_exec_id": execution_id},
        timeout=timeout_seconds,
    ) as stream_response:
        stream_response.raise_for_status()
        async for line in stream_response.aiter_lines():
            if not line or line.startswith(":"):
                continue
            if line.startswith(SSE_DATA_PREFIX):
                event_data = line[len(SSE_DATA_PREFIX) :].strip()
                if event_data:
                    event = json.loads(event_data)
                    sse_payload_adapter.validate_python(event)
                    received_events.append(event)
                    event_type = event.get("data", {}).get("event_type")
                    if event_type == WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED:
                        break
    return received_events


def get_event_types(events: list[dict[str, Any]]) -> list[str]:
    return [e.get("data", {}).get("event_type") for e in events]


def filter_events_by_type(events: list[dict[str, Any]], event_types: list[WorkflowEventType]) -> list[dict[str, Any]]:
    filtered = []
    for e in events:
        event_type = e.get("data", {}).get("event_type")
        if event_type in event_types:
            workflow_event_adapter.validate_python(e.get("data", {}))
            filtered.append(e)
    return filtered


def filter_events_by_custom_task_type(events: list[dict[str, Any]], custom_task_type: str) -> list[dict[str, Any]]:
    filtered = []
    for e in events:
        attrs = e.get("data", {}).get("attributes", {})
        if attrs.get("custom_task_type") == custom_task_type:
            workflow_event_adapter.validate_python(e.get("data", {}))
            filtered.append(e)
    return filtered


def extract_state_from_event(event: dict[str, Any]) -> dict[str, Any] | None:
    validated = sse_payload_adapter.validate_python(event)
    event_data = validated.data
    if isinstance(event_data, (CustomTaskStarted, CustomTaskInProgress, CustomTaskCompleted)):
        payload = event_data.attributes.payload
        if payload.type == "json":
            return cast(dict[str, Any], payload.value)
    return None


def extract_json_patches_from_event(event: dict[str, Any]) -> list[dict[str, Any]]:
    validated = sse_payload_adapter.validate_python(event)
    event_data = validated.data
    if isinstance(event_data, (CustomTaskInProgress, CustomTaskCompleted)):
        payload = event_data.attributes.payload
        if payload.type == "json_patch":
            return [p.model_dump() for p in payload.value]
    return []


async def execute_workflow_and_wait(
    client: httpx.AsyncClient,
    workflow_name: str,
    input_data: dict[str, Any],
    timeout_seconds: int = 30,
) -> dict[str, Any]:
    response = await client.post(
        f"/v1/workflows/{workflow_name}/execute",
        json={"input": input_data},
    )
    response.raise_for_status()
    data = response.json()

    execution_id = data.get("execution_id")
    if not execution_id:
        raise ValueError(f"No execution_id returned: {data}")

    for _ in range(timeout_seconds):
        await asyncio.sleep(1)

        status_response = await client.get(f"/v1/workflows/executions/{execution_id}")
        status_response.raise_for_status()
        status_data = status_response.json()

        status = status_data.get("status")
        if status == "COMPLETED":
            return cast(dict[str, Any], status_data)
        elif status in ["FAILED", "TERMINATED", "TIMED_OUT"]:
            raise RuntimeError(f"Workflow {workflow_name} ended with status: {status}")

    raise TimeoutError(f"Workflow {workflow_name} did not complete within {timeout_seconds}s")



import asyncio
import http
import uuid
from typing import Any, AsyncGenerator, Dict, TypeVar

import httpx
import orjson
import structlog
from pydantic import BaseModel

from mistralai_workflows.core.config.config import config
from mistralai_workflows.core.encoding.payload_encoder import PayloadEncoder
from mistralai_workflows.core.utils.id_generator import generate_two_part_id
from mistralai_workflows.exceptions import ErrorCode, WorkflowsException
from mistralai_workflows.models import (
    NetworkEncodedInput,
    ScheduleDefinition,
    WorkflowContext,
    WorkflowSpecWithTaskQueue,
)
from mistralai_workflows.protocol.v1.events import (
    WorkflowEvent,
    WorkflowEventRequest,
    WorkflowEventResponse,
)
from mistralai_workflows.protocol.v1.streaming import (
    PublishStreamEventRequest,
    PublishStreamEventResponse,
    StreamEvent,
    StreamEventsQueryParams,
)
from mistralai_workflows.protocol.v1.worker import WorkerInfo
from mistralai_workflows.protocol.v1.workflow import (
    QueryInvocationBody,
    QueryWorkflowResponse,
    ResetInvocationBody,
    SignalInvocationBody,
    SignalWorkflowResponse,
    TerminateWorkflowRequest,
    UpdateInvocationBody,
    UpdateWorkflowResponse,
    WorkflowExecutionListResponse,
    WorkflowExecutionRequest,
    WorkflowExecutionResponse,
    WorkflowExecutionStatus,
    WorkflowExecutionSyncResponse,
    WorkflowExecutionTraceEventsResponse,
    WorkflowExecutionTraceSummaryResponse,
    WorkflowExecutionTracOTelResponse,
    WorkflowGetResponse,
    WorkflowListResponse,
    WorkflowScheduleListResponse,
    WorkflowScheduleRequest,
    WorkflowScheduleResponse,
    WorkflowSpecsRegisterRequest,
    WorkflowSpecsRegisterResponse,
    WorkflowVersionListResponse,
)

logger = structlog.get_logger(__name__)

ResponseType = TypeVar(
    "ResponseType",
    WorkflowExecutionResponse,
    WorkflowExecutionSyncResponse,
    QueryWorkflowResponse,
    UpdateWorkflowResponse,
)

SCHEDULE_CORRELATION_ID_PLACEHOLDER = "__scheduled_workflow__"


class WorkflowsClient:
    """Client for interacting with the Workflows API."""

    _scheduler_namespace: str | None = None

    def __init__(
        self,
        base_url: str,
        api_version: str = "v1",
        timeout: float = 60.0,
        api_key: str | None = None,
        headers: dict[str, str] | None = None,
    ):
        """
        Args:
            base_url: The base URL of the Workflows API.
            api_version: The API version to use.
            timeout: The timeout for API requests in seconds.
            api_key: Optional API key for authentication (sent as Bearer token).
        """
        self.base_url = base_url.rstrip("/")
        self.api_version = api_version
        self.timeout = timeout
        self.api_key = api_key

        self.payload_encoder = PayloadEncoder(
            offloading_config=config.worker.temporal_payload_offloading,
            encryption_config=config.worker.temporal_payload_encryption,
        )
        if headers is None:
            headers = {}
        if api_key:
            headers["Authorization"] = f"Bearer {api_key}"

        self.client = httpx.AsyncClient(timeout=timeout, verify=config.common.ca_bundle or True, headers=headers)
        self.api_url = f"{self.base_url}/{self.api_version}/workflows"
        logger.debug(
            "Initialized WorkflowsClient",
            api_url=self.api_url,
            timeout=self.timeout,
            api_key=api_key,
        )

    async def close(self) -> None:
        """Close the underlying HTTP client."""
        await self.client.aclose()

    async def __aenter__(self) -> "WorkflowsClient":
        return self

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        await self.close()

    def _serialize_data_to_json(self, data: Any | None) -> bytes | None:
        """Serialize data to JSON bytes using orjson, with Pydantic support."""
        if data is None:
            return None

        def orjson_default(obj: Any) -> Any:
            if isinstance(obj, BaseModel):
                return obj.model_dump(mode="json")
            raise WorkflowsException(
                code=ErrorCode.UNSERIALIZABLE_PAYLOAD_ERROR,
                message=f"Type {type(obj).__name__} not serializable",
                status=http.HTTPStatus.UNPROCESSABLE_ENTITY,
            )

        if isinstance(data, BaseModel):
            python_data = data.model_dump(serialize_as_any=True)
            return orjson.dumps(python_data, default=orjson_default)
        else:
            return orjson.dumps(data, default=orjson_default)

    async def _request(
        self,
        method: str,
        endpoint: str,
        params: Dict[str, Any] | None = None,
        data: Dict[str, Any] | BaseModel | None = None,
    ) -> httpx.Response:
        """Make a request to the Workflows API.

        Args:
            method: The HTTP method to use.
            endpoint: The API endpoint to request.
            params: Optional query parameters.
            data: Optional request body.

        Returns:
            The HTTP response.

        Raises:
            httpx.HTTPStatusError: If the request fails.
        """
        url = f"{self.api_url}{endpoint}"

        # Convert Pydantic models to dict
        json_data: bytes | None = self._serialize_data_to_json(data)

        logger.debug(
            "Making request to Workflows API",
            method=method,
            url=url,
            params=params,
            json_data=json_data,
        )

        if method.upper() == "GET":
            response = await self.client.get(url, params=params)
        elif method.upper() == "POST":
            response = await self.client.post(
                url, params=params, content=json_data, headers={"Content-Type": "application/json"}
            )
        elif method.upper() == "DELETE":
            response = await self.client.delete(url, params=params)
        else:
            raise ValueError(f"Unsupported HTTP method: {method}")

        if response.status_code >= 400:
            logger.error(response.content)

        response.raise_for_status()

        return response

    async def _decode_response(self, response: ResponseType) -> ResponseType:
        if response.result and self.payload_encoder and self.payload_encoder.check_is_payload_encoded(response.result):
            response.result = await self.payload_encoder.decode_network_result(response.result)
        return response

    async def publish_stream_event(self, event: StreamEvent) -> int:
        try:
            response = await self._request(
                "POST", "/executions/stream", data=PublishStreamEventRequest(event=event).model_dump()
            )
            response.raise_for_status()
            ack = PublishStreamEventResponse.model_validate(response.json())
            return ack.broker_sequence
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to publish stream event", code=ErrorCode.POST_EXECUTIONS_STREAM_ERROR
            ) from exc

    async def stream_events(self, params: StreamEventsQueryParams) -> AsyncGenerator[StreamEvent, None]:
        """Stream workflow/activity events as SSE and yield parsed events."""
        endpoint = "/events/stream"
        query = params.model_dump(exclude_none=True)
        try:
            async with self.client.stream("GET", f"{self.api_url}{endpoint}", params=query) as response:
                response.raise_for_status()

                event_type: str | None = None
                async for line in response.aiter_lines():
                    if line is None or line == "" or line.startswith(":"):
                        continue

                    if line.startswith("event:"):
                        event_type = line[len("event:") :].strip()
                        continue

                    if line.startswith("data:"):
                        raw_data = line[len("data:") :].strip()
                        try:
                            parsed = orjson.loads(raw_data)

                            # Handle error events from server
                            if event_type == "error" or (isinstance(parsed, dict) and "error" in parsed):
                                if isinstance(parsed, dict):
                                    error_msg = parsed.get("error", "Unknown stream error")
                                else:
                                    error_msg = str(parsed)
                                raise WorkflowsException(
                                    message=f"Stream error from server: {error_msg}",
                                    code=ErrorCode.GET_EVENTS_STREAM_ERROR,
                                )

                            yield StreamEvent.model_validate(parsed)
                        except WorkflowsException:
                            raise
                        except Exception as e:
                            logger.warning("Failed to parse SSE payload", error=str(e), raw_data=raw_data)
                        finally:
                            event_type = None  # Reset for next event
        except WorkflowsException:
            raise
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to stream events", code=ErrorCode.GET_EVENTS_STREAM_ERROR
            ) from exc

    async def get_workflows(
        self,
        active_only: bool = True,
        limit: int | None = None,
        cursor: uuid.UUID | None = None,
    ) -> WorkflowListResponse:
        """Get workflows."""
        params: dict[str, Any] = {}
        if active_only:
            params["active_only"] = active_only
        if limit is not None:
            params["limit"] = limit
        if cursor is not None:
            params["cursor"] = cursor
        try:
            response = await self._request("GET", "", params=params)
            return WorkflowListResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get workflows", code=ErrorCode.GET_WORKFLOWS_ERROR
            ) from exc

    async def get_workflow_versions(
        self,
        workflow_name: str | None = None,
        workflow_id: uuid.UUID | None = None,
        active_only: bool = True,
        with_workflow: bool = False,
        limit: int | None = None,
        cursor: uuid.UUID | None = None,
    ) -> WorkflowVersionListResponse:
        """List workflow versions."""
        params: dict[str, Any] = {}
        if workflow_name:
            params["workflow_name"] = workflow_name
        if workflow_id:
            params["workflow_id"] = workflow_id
        if active_only:
            params["active_only"] = active_only
        if with_workflow:
            params["with_workflow"] = with_workflow
        if limit is not None:
            params["limit"] = limit
        if cursor is not None:
            params["cursor"] = cursor
        try:
            response = await self._request("GET", "/versions", params=params)
            return WorkflowVersionListResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get workflow versions", code=ErrorCode.GET_WORKFLOWS_VERSIONS_ERROR
            ) from exc

    async def get_scheduler_namespace(self) -> str:
        try:
            if not self._scheduler_namespace:
                self._scheduler_namespace = (await self.who_am_i()).namespace
            return self._scheduler_namespace
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get scheduler namespace", code=ErrorCode.GET_WORKERS_WHOAMI_ERROR
            ) from exc

    async def who_am_i(self) -> WorkerInfo:
        try:
            response = await self._request("GET", "/workers/whoami")
            return WorkerInfo.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get worker info", code=ErrorCode.GET_WORKERS_WHOAMI_ERROR
            ) from exc

    async def get_workflow(self, workflow_identifier: str | uuid.UUID) -> WorkflowGetResponse:
        """Get a workflow definition by name or ID."""
        try:
            response = await self._request("GET", f"/{workflow_identifier}")
            return WorkflowGetResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get workflow", code=ErrorCode.GET_WORKFLOWS_ERROR
            ) from exc

    async def register_workflow_specs(
        self,
        definitions: list[WorkflowSpecWithTaskQueue],
    ) -> WorkflowSpecsRegisterResponse:
        """Register workflow specs.

        Args:
            definitions: The workflow specs to register.
        """
        try:
            request = WorkflowSpecsRegisterRequest(definitions=definitions)
            response = await self._request("POST", "/register", data=request)
            return WorkflowSpecsRegisterResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to register workflow specs", code=ErrorCode.POST_WORKFLOWS_REGISTER_ERROR
            ) from exc

    async def execute_workflow(
        self,
        workflow_identifier: str,
        input_data: BaseModel | None,
        execution_id: str | None = None,
        task_queue: str | None = None,
        wait_for_result: bool = False,
        timeout_seconds: float | None = None,
        custom_tracing_attributes: dict[str, str] | None = None,
    ) -> WorkflowExecutionResponse | WorkflowExecutionSyncResponse:
        """Execute a workflow.

        Args:
            workflow_identifier: The workflow name or ID.
            execution_request: The execution request.

        Returns:
            The workflow execution response. If execution_request.wait_for_result is True,
            returns a WorkflowExecutionSyncResponse with the result, otherwise returns
            a WorkflowExecutionResponse.
        """
        try:
            if not execution_id:
                execution_id = generate_two_part_id(workflow_identifier)

            input_data_dict = input_data.model_dump(mode="json") if input_data else {}

            encoded_input = await self._encode_input_if_needed(execution_id=execution_id, input_data=input_data_dict)
            response = await self._request(
                "POST",
                f"/{workflow_identifier}/execute",
                data=WorkflowExecutionRequest(
                    execution_id=execution_id,
                    wait_for_result=wait_for_result,
                    timeout_seconds=timeout_seconds,
                    input=encoded_input,
                    custom_tracing_attributes=custom_tracing_attributes,
                    task_queue=task_queue,
                ),
            )
            response_data = response.json()

            # Check if this is a synchronous response by looking for wait_for_result
            if wait_for_result:
                return await self._decode_response(WorkflowExecutionSyncResponse.model_validate(response_data))
            else:
                return await self._decode_response(WorkflowExecutionResponse.model_validate(response_data))
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to execute workflow", code=ErrorCode.POST_WORKFLOWS_EXECUTE_ERROR
            ) from exc

    async def terminate_workflow_execution(
        self,
        execution_id: str,
    ) -> None:
        """Execute a workflow.

        Args:
            execution_id: The id of the workflow execution to terminate.

        Returns:
            None
        """
        try:
            response = await self._request(
                "POST",
                f"/executions/{execution_id}/terminate",
                data=TerminateWorkflowRequest(
                    execution_id=execution_id,
                ),
            )
            assert response.status_code == 204
            return
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to terminate workflow execution", code=ErrorCode.POST_EXECUTIONS_TERMINATE_ERROR
            ) from exc

    async def reset_workflow(
        self,
        execution_id: str,
        event_id: int,
        reason: str | None = None,
        exclude_signals: bool = False,
        exclude_updates: bool = False,
    ) -> None:
        """Reset a workflow execution to a specific event ID.

        Args:
            execution_id: The ID of the workflow execution to reset.
            event_id: The event ID to reset to.
            reason: Optional reason for the reset.
            exclude_signals: Whether to exclude signals that happened after the reset point.
            exclude_updates: Whether to exclude updates that happened after the reset point.

        Returns:
            None
        """
        try:
            response = await self._request(
                "POST",
                f"/executions/{execution_id}/reset",
                data=ResetInvocationBody(
                    event_id=event_id,
                    reason=reason,
                    exclude_signals=exclude_signals,
                    exclude_updates=exclude_updates,
                ),
            )
            assert response.status_code == 204
            return
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to reset workflow execution", code=ErrorCode.TEMPORAL_SERVICE_ERROR
            ) from exc

    async def get_workflow_executions(
        self,
        workflow_identifier: str | uuid.UUID | None = None,
        page_size: int = 50,
        next_page_token: str | None = None,
    ) -> WorkflowExecutionListResponse:
        """Get all workflow executions with token-based pagination.

        Args:
            page_size: Number of items per page
            next_page_token: Token for the next page of results

        Returns:
            A list of workflow executions with pagination information.
        """
        params: Dict[str, Any] = {"page_size": page_size}
        if workflow_identifier:
            params["workflow_identifier"] = workflow_identifier
        if next_page_token:
            params["next_page_token"] = next_page_token
        endpoint = "/executions"

        try:
            response = await self._request("GET", endpoint, params=params)
            return WorkflowExecutionListResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get workflow executions", code=ErrorCode.GET_EXECUTIONS_ERROR
            ) from exc

    async def get_workflow_execution(self, execution_id: str) -> WorkflowExecutionResponse:
        """Get a workflow execution by ID.

        Args:
            execution_id: The ID of the workflow execution.

        Returns:
            The workflow execution.
        """
        try:
            response = await self._request("GET", f"/executions/{execution_id}")
            return await self._decode_response(WorkflowExecutionResponse.model_validate(response.json()))
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get workflow execution", code=ErrorCode.GET_EXECUTIONS_ERROR
            ) from exc

    async def get_workflow_execution_trace_otel(self, execution_id: str) -> WorkflowExecutionTracOTelResponse:
        """Get the OpenTelemetry trace data for a workflow execution.

        Args:
            execution_id: The ID of the workflow execution.

        Returns:
            The OpenTelemetry trace data.
        """
        try:
            response = await self._request("GET", f"/executions/{execution_id}/trace/otel")
            return WorkflowExecutionTracOTelResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc,
                message="Failed to get workflow execution trace otel",
                code=ErrorCode.GET_EXECUTIONS_TRACE_OTEL_ERROR,
            ) from exc

    async def get_workflow_execution_trace_summary(self, execution_id: str) -> WorkflowExecutionTraceSummaryResponse:
        """Get the trace summary for a workflow execution.

        Args:
            execution_id: The ID of the workflow execution.

        Returns:
            The trace summary.
        """
        try:
            response = await self._request("GET", f"/executions/{execution_id}/trace/summary")
            return WorkflowExecutionTraceSummaryResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc,
                message="Failed to get workflow execution trace summary",
                code=ErrorCode.GET_EXECUTIONS_TRACE_SUMMARY_ERROR,
            ) from exc

    async def get_workflow_execution_trace_events(
        self,
        execution_id: str,
        merge_same_id_events: bool = False,
        include_internal_events: bool = False,
    ) -> WorkflowExecutionTraceEventsResponse:
        """Fetch trace events for a specific workflow execution.

        Args:
            execution_id: The ID of the workflow execution to fetch events for.
            merge_same_id_events: If True, merges events with the same ID, retaining the last one.
                                  Useful for UI progress display to show the latest event status.
            include_internal_events: If True, includes internal events in the response.

        Returns:
            WorkflowExecutionTraceEventsResponse: An object containing the trace events.
        """
        try:
            response = await self._request(
                "GET",
                f"/executions/{execution_id}/trace/events",
                params={
                    "merge_same_id_events": merge_same_id_events,
                    "include_internal_events": include_internal_events,
                },
            )
            return WorkflowExecutionTraceEventsResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc,
                message="Failed to get workflow execution trace events",
                code=ErrorCode.GET_EXECUTIONS_TRACE_EVENTS_ERROR,
            ) from exc

    async def wait_for_workflow_completion(
        self,
        execution_id: str,
        polling_interval: int = 5,
        max_attempts: int | None = None,
    ) -> WorkflowExecutionResponse:
        """Wait for a workflow to complete by polling its status.

        Args:
            execution_id: Execution ID of the workflow
            polling_interval: Seconds between status checks
            max_attempts: Maximum number of polling attempts (None for unlimited)

        Returns:
            WorkflowExecutionResponse with the final execution details

        Raises:
            TimeoutError: If max_attempts is reached and workflow is still running
            RuntimeError: If workflow fails or terminates abnormally
        """
        attempts = 0
        try:
            while True:
                response = await self.get_workflow_execution(execution_id)

                if response.status != WorkflowExecutionStatus.RUNNING:
                    if response.status == WorkflowExecutionStatus.COMPLETED:
                        return response
                    else:
                        raise WorkflowsException(
                            message=f"Workflow failed with status: {response.status}",
                            code=ErrorCode.GET_EXECUTIONS_ERROR,
                        )

                attempts += 1
                if max_attempts is not None and attempts >= max_attempts:
                    raise WorkflowsException(
                        message=f"Workflow is still running after {max_attempts} polling attempts",
                        code=ErrorCode.GET_EXECUTIONS_ERROR,
                    )

                await asyncio.sleep(polling_interval)
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed while waiting for workflow completion", code=ErrorCode.GET_EXECUTIONS_ERROR
            ) from exc

    async def execute_workflow_and_wait(
        self,
        workflow_identifier: str,
        input_data: BaseModel | None,
        task_queue: str | None = None,
        execution_id: str | None = None,
        polling_interval: int = 5,
        max_attempts: int | None = None,
        use_api_sync: bool = False,
        timeout_seconds: float | None = None,
        custom_tracing_attributes: Dict[str, str] | None = None,
    ) -> Any:
        """Execute a workflow and wait for its completion.

        Args:
            workflow_identifier: The workflow name or ID.
            input_data: Input parameters for the workflow
            execution_id: Optional custom execution ID
            polling_interval: Seconds between status checks when polling
            max_attempts: Maximum number of polling attempts when polling (None for unlimited)
            use_api_sync: Whether to use the API's built-in sync execution capability
            timeout_seconds: Maximum time to wait in seconds when using API sync

        Returns:
            The workflow result directly

        Raises:
            TimeoutError: If max_attempts is reached and workflow is still running
            RuntimeError: If workflow fails or terminates abnormally
            httpx.HTTPStatusError: If API request fails (including timeouts)
        """
        try:
            if use_api_sync:
                # Use the API's built-in synchronous execution
                response = await self.execute_workflow(
                    workflow_identifier,
                    input_data=input_data,
                    execution_id=execution_id,
                    wait_for_result=True,
                    timeout_seconds=timeout_seconds,
                    custom_tracing_attributes=custom_tracing_attributes,
                    task_queue=task_queue,
                )
                return response.result
            else:
                # Use polling method
                execution = await self.execute_workflow(
                    workflow_identifier,
                    input_data=input_data,
                    execution_id=execution_id,
                    custom_tracing_attributes=custom_tracing_attributes,
                    task_queue=task_queue,
                )

                # Wait for completion
                final_execution = await self.wait_for_workflow_completion(
                    execution.execution_id, polling_interval, max_attempts
                )

                return final_execution.result
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to execute workflow and wait", code=ErrorCode.POST_WORKFLOWS_EXECUTE_ERROR
            ) from exc

    async def _encode_input_if_needed(
        self,
        execution_id: str,
        input_data: Dict[str, Any],
    ) -> Dict[str, Any] | Any | NetworkEncodedInput | None:
        if not self.payload_encoder:
            return input_data
        encoded_input = await self.payload_encoder.encode_network_input(
            input_data,
            context=WorkflowContext(
                namespace=await self.get_scheduler_namespace(),
                execution_id=execution_id,
            ),
        )
        return encoded_input

    async def signal_workflow(
        self,
        execution_id: str,
        signal_name: str,
        input_data: BaseModel | None = None,
    ) -> SignalWorkflowResponse:
        """Send a signal to a running workflow execution."""
        endpoint = f"/executions/{execution_id}/signals"
        try:
            input_data_dict = input_data.model_dump(mode="json") if input_data else {}
            encoded_input = await self._encode_input_if_needed(execution_id, input_data_dict)
            request_body = SignalInvocationBody(name=signal_name, input=encoded_input)
            logger.debug(
                "WorkflowsClient: Sending signal",
                execution_id=execution_id,
                signal_name=signal_name,
            )
            response = await self._request("POST", endpoint, data=request_body)
            return SignalWorkflowResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to signal workflow", code=ErrorCode.POST_EXECUTIONS_SIGNALS_ERROR
            ) from exc

    async def query_workflow(
        self,
        execution_id: str,
        query_name: str,
        input_data: BaseModel | None = None,
    ) -> QueryWorkflowResponse:
        """Query a running workflow execution."""
        endpoint = f"/executions/{execution_id}/queries"
        try:
            input_data_dict = input_data.model_dump(mode="json") if input_data else {}
            encoded_input = await self._encode_input_if_needed(execution_id, input_data_dict)
            request_body = QueryInvocationBody(name=query_name, input=encoded_input)
            logger.debug(
                "WorkflowsClient: Sending query",
                execution_id=execution_id,
                query_name=query_name,
            )
            response = await self._request("POST", endpoint, data=request_body)
            return await self._decode_response(QueryWorkflowResponse.model_validate(response.json()))
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to query workflow", code=ErrorCode.POST_EXECUTIONS_QUERIES_ERROR
            ) from exc

    async def update_workflow(
        self,
        execution_id: str,
        update_name: str,
        input_data: BaseModel | None = None,
    ) -> UpdateWorkflowResponse:
        """Send an update to a running workflow execution.

        Args:
            execution_id (str): The execution ID of the workflow to update.
            update_name (str): The name of the update to send.
            input_data (Dict[str, Any] | None, optional): The input data for the update. Defaults to None.

        Returns:
            UpdateWorkflowResponse: The response from the server.
        """
        endpoint = f"/executions/{execution_id}/updates"
        try:
            input_data_dict = input_data.model_dump(mode="json") if input_data else {}

            encoded_input = await self._encode_input_if_needed(execution_id, input_data_dict)
            request_body = UpdateInvocationBody(name=update_name, input=encoded_input)
            logger.debug("WorkflowsClient: Sending update", execution_id=execution_id, update_name=update_name)
            response = await self._request("POST", endpoint, data=request_body)
            return await self._decode_response(UpdateWorkflowResponse.model_validate(response.json()))
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to update workflow", code=ErrorCode.POST_EXECUTIONS_UPDATES_ERROR
            ) from exc

    async def schedule_workflow(
        self,
        schedule: ScheduleDefinition,
        schedule_id: str | None = None,
        workflow_version_id: uuid.UUID | None = None,
        workflow_identifier: str | None = None,
        workflow_task_queue: str | None = None,
    ) -> WorkflowScheduleResponse:
        """Schedule a workflow execution. Either by name, or version id"""
        try:
            if not workflow_version_id and not workflow_identifier:
                raise WorkflowsException(
                    message="Either workflow_version_id or workflow_identifier must be provided",
                    code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                )
            elif workflow_version_id and workflow_identifier:
                raise WorkflowsException(
                    message="Only one of workflow_version_id or workflow_identifier can be provided",
                    code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                )

            if workflow_identifier and not workflow_task_queue:
                # Fallback on default task queue:
                workflow_task_queue = config.temporal.task_queue

            endpoint = "/schedules"

            schedule.input = await self._encode_input_if_needed(
                SCHEDULE_CORRELATION_ID_PLACEHOLDER,
                schedule.input,
            )

            request_body = WorkflowScheduleRequest(
                schedule=schedule,
                schedule_id=schedule_id,
                workflow_version_id=workflow_version_id,
                workflow_identifier=workflow_identifier,
                workflow_task_queue=workflow_task_queue,
            )

            response = await self._request("POST", endpoint, data=request_body)
            return WorkflowScheduleResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to schedule workflow", code=ErrorCode.POST_SCHEDULES_ERROR
            ) from exc

    async def unschedule_workflow(self, schedule_id: str) -> None:
        """Unschedule a workflow execution."""
        try:
            await self._request("DELETE", f"/schedules/{schedule_id}")
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to unschedule workflow", code=ErrorCode.DELETE_SCHEDULES_ERROR
            ) from exc

    async def get_schedules(self) -> WorkflowScheduleListResponse:
        """Get all workflow schedules."""
        try:
            response = await self._request("GET", "/schedules")
            return WorkflowScheduleListResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to get schedules", code=ErrorCode.GET_SCHEDULES_ERROR
            ) from exc

    async def send_event(self, event: WorkflowEvent) -> WorkflowEventResponse:
        """Send a workflow event to the API.

        Args:
            event: The workflow event to send.

        Returns:
            The response from the server.
        """
        try:
            response = await self._request(
                "POST",
                "/events",
                data=WorkflowEventRequest(event=event),
            )
            return WorkflowEventResponse.model_validate(response.json())
        except Exception as exc:
            raise WorkflowsException.from_api_client_error(
                exc, message="Failed to send workflow event", code=ErrorCode.POST_EVENTS_ERROR
            ) from exc



import abc
from typing import Generic, TypeVar

import mistralai
from pydantic import BaseModel

from mistralai_workflows.plugins.agents.agent import Agent

T = TypeVar("T", bound=BaseModel)
Inputs = list[T | str | mistralai.ContentChunk]
Outputs = list[T]
FinalOutputs = list[mistralai.ContentChunk]


class Session(abc.ABC, Generic[T]):
    @abc.abstractmethod
    def is_conversation_active(self) -> bool: ...

    @abc.abstractmethod
    async def initialize_conversation(self, agent: Agent, inputs: Inputs[T]) -> Outputs[T]: ...

    @abc.abstractmethod
    async def close_conversation(self) -> None: ...

    @abc.abstractmethod
    async def append_messages(self, inputs: Inputs[T]) -> Outputs[T]: ...

    @abc.abstractmethod
    async def process_output(self, output: T) -> Outputs[T]: ...

    @abc.abstractmethod
    def format_final_outputs(self, outputs: Outputs[T]) -> FinalOutputs: ...



"""
Monitoring and logging configuration utilities.

This module provides structured logging setup using structlog.
"""

import logging
from enum import StrEnum
from typing import Any

import structlog
from opentelemetry import trace
from pydantic import BaseModel
from structlog.typing import EventDict, Processor


class Env(StrEnum):
    DEV = "development"
    STAGING = "staging"
    PROD = "production"


class LogFormat(StrEnum):
    JSON = "json"
    CONSOLE = "console"


class LogLevel(StrEnum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class LoggerConfig(BaseModel):
    name: str
    level: LogLevel


def _get_log_level_value(level: LogLevel) -> int:
    level_map = {
        LogLevel.DEBUG: logging.DEBUG,
        LogLevel.INFO: logging.INFO,
        LogLevel.WARNING: logging.WARNING,
        LogLevel.ERROR: logging.ERROR,
        LogLevel.CRITICAL: logging.CRITICAL,
    }
    return level_map[level]


def _add_otel_trace_processor(logger: Any, method_name: str, event_dict: EventDict) -> EventDict:
    span = trace.get_current_span()
    if span:
        ctx = span.get_span_context()
        if ctx.is_valid:
            event_dict["otel.trace_id"] = format(ctx.trace_id, "032x")
            event_dict["otel.span_id"] = format(ctx.span_id, "016x")

    return event_dict


def setup_logging(
    log_level: LogLevel = LogLevel.INFO,
    log_format: LogFormat = LogFormat.CONSOLE,
    app_version: str = "0.0.0",
    inject_otel_trace: bool = False,
    extra_config: list[LoggerConfig] | None = None,
) -> None:
    logging.basicConfig(
        format="%(message)s",
        level=_get_log_level_value(log_level),
    )

    if extra_config:
        for logger_config in extra_config:
            logger = logging.getLogger(logger_config.name)
            logger.setLevel(_get_log_level_value(logger_config.level))

    processors: list[Processor] = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso"),
        structlog.processors.StackInfoRenderer(),
    ]

    if inject_otel_trace:
        processors.append(_add_otel_trace_processor)

    processors.append(
        structlog.processors.CallsiteParameterAdder(
            {
                structlog.processors.CallsiteParameter.PATHNAME,
                structlog.processors.CallsiteParameter.LINENO,
            }
        )
    )

    if log_format == LogFormat.JSON:
        processors.extend(
            [
                structlog.processors.format_exc_info,
                structlog.processors.JSONRenderer(),
            ]
        )
    else:
        processors.extend(
            [
                structlog.processors.format_exc_info,
                structlog.dev.ConsoleRenderer(),
            ]
        )

    structlog.configure(
        processors=processors,
        wrapper_class=structlog.stdlib.BoundLogger,
        context_class=dict,
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )



from typing import Any, Dict

from pydantic import BaseModel, Field


class SignalDefinition(BaseModel):
    name: str = Field(description="Name of the signal")
    description: str | None = Field(default=None, description="Description of the signal")
    input_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Input JSON schema of the signal's model",
        json_schema_extra={"additionalProperties": True},
    )
    # Signals don't have an output schema from the sender's perspective


class QueryDefinition(BaseModel):
    name: str = Field(description="Name of the query")
    description: str | None = Field(default=None, description="Description of the query")
    input_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Input JSON schema of the query's model",
        json_schema_extra={"additionalProperties": True},
    )
    output_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Output JSON schema of the query's model",
        json_schema_extra={"additionalProperties": True},
    )


class UpdateDefinition(BaseModel):
    name: str = Field(description="Name of the update")
    description: str | None = Field(default=None, description="Description of the update")
    input_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Input JSON schema of the update's model",
        json_schema_extra={"additionalProperties": True},
    )
    output_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Output JSON schema of the update's model",
        json_schema_extra={"additionalProperties": True},
    )



# ruff: noqa: E501
# Why `LocalSession` was built (vs. `RemoteSession` aka Agents SDK)?
# To address some temporary limitations of the Agents SDK, `LocalSession` was created to provide more control and
# flexibility. This is mostly experimental and will probably be removed in the future once the Agents SDK has the
# missing features.
#
# - [x] On-premises compatibility:
#   Agents SDK is not yet available for on-prem deployments; `LocalSession` operates independently, using only the
#   completion endpoint.
#
# - [ ] Full context control:
#   Zero-latency direct manipulation of `messages` list (no API roundtrips). Supports granular operations (e.g.,
#   selective message removal) not natively available in the Agents SDK.
#
# - [ ] Ad-hoc model compatibility:
#   Agents SDK rely on Mistral-specific features (e.g., `prompt_mode`, `prompt_data`), limiting compatibility with
#   non-Mistral models; `LocalSession` uses standard function-calling interfaces for broader experimentation.
#   References:
#   - https://github.com/mistralai/mistral/blob/8e69df1dcb826dbe4e9f8582f87f22282a604b72/common/mistral_common/protocol/system_prompt/templates/shared/additional_integrations.v2_6_8.jinja2#L22-L23
#   - https://github.com/mistralai/dashboard/blob/ea41933dd1705dfaa9cd81ceca6d394fee8400e1/bora/bora/services/completion.py#L532-L539
#
# - [ ] Full customization:
#   Direct access to system prompts and execution logic allowing full optimization of the full agent system.

import asyncio
import re
import uuid
from typing import cast

import mistralai
import structlog
from mistralai_workflows.plugins.mistralai.activities import mistralai_chat_complete, mistralai_chat_stream

from mistralai_workflows.plugins.agents.agent import Agent
from mistralai_workflows.plugins.agents.session.session import FinalOutputs, Inputs, Outputs, Session
from mistralai_workflows.plugins.agents.tool import (
    check_is_custom_tool,
    convert_tool_to_mistral_tool,
    execute_activity_tool,
    get_tool_name,
)

logger = structlog.get_logger(__name__)

Message = mistralai.Messages
LocalSessionInputs = Inputs[Message]
LocalSessionOutputs = Outputs[Message]


# inspired from https://github.com/mistralai/dashboard/blob/6d3dea5d527518e68b9fd159671e27a714437811/bora/bora/tools/handoff.py#L25
# ruff: noqa: W291, E501
SYSTEM_PROMPT_AGENT_HANDOFF = """
You are part of a multi-agent system, designed to make agent coordination and execution easy. An agent encompasses instructions and tools and can hand off a conversation to another agent when appropriate. Handoffs are achieved by calling a handoff function named `handoff_to_<agent_name>` with empty params (`{{}}`). Transfers between agents are handled seamlessly in the background; do not mention or draw attention to these transfers in your conversation with the user.

### When to transfer to another agent

When you don't have the capabilities or available tools to answer the request you can pick another agent that is better suited to add new information to the request.

### When to not transfer to another agent

If you are capable of answering the request, don't transfer it on to another agent and complete the request. Following are the available handoff tools:
{AGENT_HANDOFF_TOOLS}
""".strip()


# inspired from https://github.com/mistralai/dashboard/blob/6d3dea5d527518e68b9fd159671e27a714437811/bora/bora/protocol/v1/agents.py#L49-L83
def _get_agent_handoff_tool_description(agent: Agent) -> str:
    tool_names = [get_tool_name(tool) for tool in agent.tools or []]
    rep = (
        f"Invokes the Agent("
        f"name={agent.name}, "
        f"description={agent.description}, "
        f"available_tools={', '.join(tool_name for tool_name in tool_names if tool_name) if agent.tools else 'None'}"
        f")"
    )
    return rep


# inspired from https://github.com/mistralai/dashboard/blob/6d3dea5d527518e68b9fd159671e27a714437811/bora/bora/protocol/v1/agents.py#L49-L83
def _get_agent_handoff_tool_name(agent: Agent) -> str:
    snake_case_name = agent.name.replace("-", "_").replace(" ", "")
    snake_case_name = re.sub(r"(?<!^)(?=[A-Z])", "_", snake_case_name).lower()
    return f"handoff_to_{snake_case_name}"


def _get_agent_to_handoff_tool(agent_handoffs: list[Agent]) -> dict[Agent, mistralai.Tool]:
    return {
        agent: mistralai.Tool(
            type="function",
            function=mistralai.Function(
                name=_get_agent_handoff_tool_name(agent),
                description=_get_agent_handoff_tool_description(agent),
                parameters={},
            ),
        )
        for agent in agent_handoffs
    }


class LocalSession(Session[Message]):
    def __init__(
        self,
        system_prompt_agent_handoff: str = SYSTEM_PROMPT_AGENT_HANDOFF,
        raise_on_tool_fail: bool = True,
        stream: bool = False,
    ) -> None:
        logger.warning(
            "You are using the experimental `LocalSession` class. We recommend using `RemoteSession` and ONLY use `LocalSession` if you hit some current limitations of the Agents SDK."
        )

        self._raise_on_tool_fail = raise_on_tool_fail
        self._system_prompt_agent_handoff = system_prompt_agent_handoff
        self._stream = stream
        self._conversation_id: str | None = None
        self._current_agent: Agent | None = None
        self._available_agents: list[Agent] = []
        self._handoff_function_name_to_agent: dict[str, Agent] = {}
        self._agent_to_handoff_tool: dict[Agent, mistralai.Tool] = {}
        self._agent_to_tools: dict[Agent, list[mistralai.Tool]] = {}
        self.messages: list[Message] = []

    def is_conversation_active(self) -> bool:
        return self._conversation_id is not None

    async def initialize_conversation(self, agent: Agent, inputs: LocalSessionInputs) -> LocalSessionOutputs:
        if self._conversation_id:
            raise ValueError("Session already started")

        self._available_agents = list(agent.iterate_agents_deeply_in_handoffs(agent))
        self._agent_to_handoff_tool = _get_agent_to_handoff_tool(self._available_agents)
        self._handoff_function_name_to_agent = self._get_handoff_function_name_to_agent(self._agent_to_handoff_tool)
        self._agent_to_tools = self._compute_agent_to_tools(self._available_agents)
        self._current_agent = agent
        self._conversation_id = str(uuid.uuid4())

        return await self.append_messages(inputs)

    async def close_conversation(self) -> None:
        self._conversation_id = None
        self._current_agent = None
        self._available_agents = []
        self._handoff_function_name_to_agent = {}
        self._agent_to_handoff_tool = {}
        self._agent_to_tools = {}
        self.messages = []

    async def append_messages(self, inputs: LocalSessionInputs) -> LocalSessionOutputs:
        if not self._conversation_id or not self._current_agent:
            raise ValueError("Session not started")

        agent = self._current_agent

        input_messages = self._convert_inputs_to_messages(inputs)
        self.messages.extend(input_messages)

        untracked_messages: list[Message] = []

        if agent.instructions:
            untracked_messages.insert(0, mistralai.SystemMessage(content=agent.instructions))

        tools = self._agent_to_tools.get(agent, []).copy()

        if self._agent_to_handoff_tool and agent.handoffs:
            assert agent in self._agent_to_handoff_tool, (
                f"Agent '{agent.name}' not found in agent_to_handoff_tool but should be"
            )
            handoff_tools = [self._agent_to_handoff_tool[agent] for agent in agent.handoffs]
            tools.extend(handoff_tools)
            system_prompt_agent_handoff = self._system_prompt_agent_handoff.format(
                AGENT_HANDOFF_TOOLS="\n".join(
                    f"- {tool.function.name}: {_get_agent_handoff_tool_description(agent)}" for tool in handoff_tools
                )
            )
            untracked_messages.insert(0, mistralai.SystemMessage(content=system_prompt_agent_handoff))

        chat_request = mistralai.ChatCompletionRequest.model_validate(
            {
                "messages": untracked_messages + self.messages,
                "model": agent.model,
                "tools": tools or mistralai.UNSET,
                **(agent.completion_args.model_dump(by_alias=True) if agent.completion_args else {}),
            }
        )

        if self._stream:
            assistant_message = await mistralai_chat_stream(chat_request)
            output_messages = [assistant_message]
        else:
            response = await mistralai_chat_complete(chat_request)
            output_messages = [choice.message for choice in response.choices]

        self.messages.extend(output_messages)

        return cast(LocalSessionOutputs, output_messages)

    async def process_output(self, output: Message) -> LocalSessionOutputs:
        if output.role != "assistant":
            # no tool calls, so no next input entries
            return []

        activity_tools: list[mistralai.ToolCall] = []
        handoff_tools: list[mistralai.ToolCall] = []

        for tool_call in output.tool_calls or []:
            if tool_call.function.name in self._handoff_function_name_to_agent:
                handoff_tools.append(tool_call)
            else:
                activity_tools.append(tool_call)

        tool_responses = await asyncio.gather(
            *[
                execute_activity_tool(tool_call.function.name, tool_call.function.arguments, self._raise_on_tool_fail)
                for tool_call in activity_tools
            ]
        )
        next_input_entries = [
            mistralai.ToolMessage(content=tool_response, tool_call_id=tool_call.id)
            for tool_call, tool_response in zip(activity_tools, tool_responses, strict=True)
        ]

        if handoff_tools:
            if len(handoff_tools) != 1:
                for tool in handoff_tools:
                    # For now we only support one handoff tool call at a time, as handoff is a function call,
                    # model is allowed to call multiple tools at once, but we don't support it yet,
                    # so we return error to model to make it them sequentially.
                    next_input_entries.append(
                        mistralai.ToolMessage(
                            content="ERROR: Only one handoff tool call is supported. Do them one by one.",
                            tool_call_id=tool.id,
                        )
                    )
            else:
                handoff_tool_name = handoff_tools[0].function.name
                assert handoff_tool_name in self._handoff_function_name_to_agent, (
                    f"Handoff tool '{handoff_tool_name}' not found but categorized as handoff tool above."
                )
                self._current_agent = self._handoff_function_name_to_agent[handoff_tool_name]
                next_input_entries.append(
                    mistralai.ToolMessage(
                        content=f"current assistant is now: {self._current_agent.name}",
                        tool_call_id=handoff_tools[0].id,
                    )
                )

        return cast(LocalSessionOutputs, next_input_entries)

    def format_final_outputs(self, outputs: LocalSessionOutputs) -> FinalOutputs:
        final_outputs: FinalOutputs = []
        for output in outputs:
            if not isinstance(output, mistralai.AssistantMessage) or output.content is None:
                continue
            if isinstance(output.content, str):
                final_outputs.append(mistralai.TextChunk(text=output.content))
            elif isinstance(output.content, list):
                final_outputs.extend(output.content)
        return final_outputs

    def _get_handoff_function_name_to_agent(
        self,
        agent_to_handoff_tool: dict[Agent, mistralai.Tool],
    ) -> dict[str, Agent]:
        return {tool.function.name: agent for agent, tool in agent_to_handoff_tool.items()}

    def _compute_agent_to_tools(self, available_agents: list[Agent]) -> dict[Agent, list[mistralai.Tool]]:
        agent_to_tools: dict[Agent, list[mistralai.Tool]] = {}
        for agent in available_agents:
            if not agent.tools:
                continue
            mistral_tools = self._get_mistralai_tools(agent)
            if not mistral_tools:
                continue
            agent_to_tools[agent] = mistral_tools
        return agent_to_tools

    def _get_mistralai_tools(self, agent: Agent) -> list[mistralai.Tool] | None:
        if not agent.tools:
            return None

        mistral_tools: list[mistralai.Tool] = []
        for tool in agent.tools:
            if not check_is_custom_tool(tool):
                logger.warning("Skipping bult-in tool when running in local mode", tool=tool)
                continue
            mistral_tool = convert_tool_to_mistral_tool(tool)
            mistral_tools.append(cast(mistralai.Tool, mistral_tool))
        return mistral_tools

    def _convert_inputs_to_messages(self, inputs: LocalSessionInputs) -> LocalSessionOutputs:
        messages: LocalSessionOutputs = []
        for inp in inputs:
            if isinstance(inp, str):
                messages.append(mistralai.UserMessage(role="user", content=inp))
            elif isinstance(
                inp,
                (mistralai.ImageURLChunk, mistralai.DocumentURLChunk, mistralai.TextChunk, mistralai.ReferenceChunk),
            ):
                messages.append(mistralai.UserMessage(role="user", content=[inp]))
            elif isinstance(
                inp, (mistralai.UserMessage, mistralai.AssistantMessage, mistralai.ToolMessage, mistralai.SystemMessage)
            ):
                messages.append(inp)
            else:
                raise ValueError(f"Unsupported input type: {type(inp)}")
        return messages



import asyncio
import inspect
from types import NoneType, UnionType
from typing import Any, Callable, Dict, Tuple, Type, TypeVar, Union, get_args, get_origin, get_type_hints

from pydantic import BaseModel, TypeAdapter, ValidationError

from mistralai_workflows.core.definition.validation.schema_generator import (
    generate_pydantic_model_from_params,
    generate_pydantic_model_from_return_type,
)
from mistralai_workflows.core.dependencies.dependency_injector import DependsCls
from mistralai_workflows.exceptions import ErrorCode, WorkflowsException

T = TypeVar("T", bound=Callable[..., Any])


def extract_origin_type(type_to_extract: Type) -> Type:
    origin = get_origin(type_to_extract)
    args: Tuple[Type, ...] = get_args(type_to_extract)
    if origin is Union and len(args) == 2 and NoneType in args:
        return next(arg for arg in args if arg is not NoneType)
    elif origin is UnionType and len(args) == 2 and NoneType in args:
        return next(arg for arg in args if arg is not NoneType)
    return type_to_extract


def check_is_valid_type(type_to_check: Type, expected_type: Type, allow_optional: bool = False) -> bool:
    if allow_optional:
        type_to_check = extract_origin_type(type_to_check)

    return (inspect.isclass(type_to_check) and issubclass(type_to_check, expected_type)) or (
        type_to_check is expected_type
    )


def get_function_signature_type_hints(func: Callable, is_method: bool) -> Tuple[Dict[str, Type], Type]:
    type_hints = get_type_hints(func)
    if "return" not in type_hints:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=(
                f"'{func.__name__}' must have a return type annotation, use `-> None` if no return type is expected."
            ),
        )

    sig = inspect.signature(func)
    param_names = list(sig.parameters.keys())

    if is_method:
        if not param_names:
            raise WorkflowsException(
                code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
                message=f"Method '{func.__name__}' has no parameters, expected 'self' or 'cls' as the first.",
            )
        type_hints.pop(param_names[0], None)
        param_names = param_names[1:]

    user_params_dict = {}
    for param_name in param_names:
        param = sig.parameters[param_name]

        if isinstance(param.default, DependsCls):
            continue

        # Only add to user params if it has a type hint
        if param_name in type_hints:
            user_params_dict[param_name] = type_hints[param_name]
        else:
            raise WorkflowsException(
                code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
                message=f"Parameter '{param_name}' in '{func.__name__}' must have a type annotation.",
            )

    return_type = type_hints["return"]
    return user_params_dict, return_type


def raise_if_function_has_invalid_signature(func: Callable, is_method: bool = False) -> None:
    if not asyncio.iscoroutinefunction(func):
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"'{func.__name__}' must be async function. Use `async def`.",
        )

    user_params_dict, return_type = get_function_signature_type_hints(func, is_method=is_method)

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"Cannot generate Pydantic model from parameters of '{func.__name__}': {e}",
        ) from e

    try:
        generate_pydantic_model_from_return_type(func.__name__, return_type)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"Cannot generate Pydantic model from return type of '{func.__name__}': {e}",
        ) from e


def raise_if_function_has_invalid_usage(
    func: Callable, args: Tuple[Any, ...], kwargs: Dict[str, Any], is_method: bool = False
) -> None:
    user_params_dict, _ = get_function_signature_type_hints(func, is_method=is_method)

    if is_method:
        args = args[1:]

    if len(kwargs) > 0:
        raise WorkflowsException(
            code=ErrorCode.INVALID_ARGUMENTS_ERROR,
            message=(
                f"'{func.__name__}' should not take keyword arguments, "
                "use positional arguments instead. Found: "
                f"{len(kwargs)} keyword arguments."
            ),
        )

    sig = inspect.signature(func)
    param_names = list(sig.parameters.keys())
    if is_method:
        param_names = param_names[1:]

    required_count = 0
    for param_name in param_names:
        param = sig.parameters[param_name]
        if isinstance(param.default, DependsCls):
            continue
        if param.default is inspect.Parameter.empty:
            required_count += 1

    total_param_count = len(user_params_dict)

    if len(args) < required_count or len(args) > total_param_count:
        raise WorkflowsException(
            code=ErrorCode.INVALID_ARGUMENTS_ERROR,
            message=(
                f"'{func.__name__}' expects {required_count} to {total_param_count} parameters. "
                f"Found: {len(args)} arguments."
            ),
        )

    for i, (param_name, param_type) in enumerate(user_params_dict.items()):
        if i < len(args):
            arg = args[i]
            if inspect.isclass(param_type) and issubclass(param_type, BaseModel):
                if not isinstance(arg, param_type):
                    raise WorkflowsException(
                        code=ErrorCode.INVALID_ARGUMENTS_ERROR,
                        message=(
                            f"Parameter '{param_name}' in '{func.__name__}' should be of type '{param_type}'. "
                            f"Found: '{type(arg)}'."
                        ),
                    )


def raise_if_function_has_invalid_return_type(func: Callable, return_value: Any, is_method: bool = False) -> None:
    _, return_type = get_function_signature_type_hints(func, is_method=is_method)

    try:
        adapter = TypeAdapter(return_type)
        adapter.validate_python(return_value, strict=True)
    except ValidationError as e:
        raise WorkflowsException(
            code=ErrorCode.ACTIVITY_DEFINITION_ERROR,
            message=f"'{func.__name__}' return type validation failed: {e}",
        ) from e


def validate_signal_handler_signature(func: Callable, is_method: bool = True) -> None:
    user_params_dict, _ = get_function_signature_type_hints(func, is_method=is_method)

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_SIGNAL_DEFINITION_ERROR,
            message=f"Signal '{func.__name__}' has invalid parameters for schema generation: {e}",
        ) from e


def validate_query_handler_signature(func: Callable, is_method: bool = True) -> None:
    if asyncio.iscoroutinefunction(func):
        raise WorkflowsException(
            code=ErrorCode.REJECTED_QUERY_ERROR,
            message=f"Query '{func.__name__}' must be a synchronous function (def), not async def.",
        )

    user_params_dict, return_type = get_function_signature_type_hints(func, is_method=is_method)

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.REJECTED_QUERY_ERROR,
            message=f"Query '{func.__name__}' has invalid parameters for schema generation: {e}",
        ) from e

    if return_type is NoneType or return_type is type(None):
        raise WorkflowsException(
            code=ErrorCode.REJECTED_QUERY_ERROR,
            message=f"Query '{func.__name__}' must have a return type annotation other than None.",
        )

    try:
        generate_pydantic_model_from_return_type(func.__name__, return_type)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.REJECTED_QUERY_ERROR,
            message=f"Query '{func.__name__}' has invalid return type for schema generation: {e}",
        ) from e


def validate_update_handler_signature(func: Callable, is_method: bool = True) -> None:
    user_params_dict, return_type = get_function_signature_type_hints(func, is_method=is_method)

    try:
        generate_pydantic_model_from_params(func.__name__, user_params_dict, func=func)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_UPDATE_DEFINITION_ERROR,
            message=f"Update '{func.__name__}' has invalid parameters for schema generation: {e}",
        ) from e

    if return_type is NoneType or return_type is type(None):
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_UPDATE_DEFINITION_ERROR,
            message=f"Update '{func.__name__}' must have a return type annotation other than None.",
        )

    try:
        generate_pydantic_model_from_return_type(func.__name__, return_type)
    except Exception as e:
        raise WorkflowsException(
            code=ErrorCode.WORKFLOW_UPDATE_DEFINITION_ERROR,
            message=f"Update '{func.__name__}' has invalid return type for schema generation: {e}",
        ) from e



"""Tests for tool utilities."""

from mistralai_workflows.plugins.agents.tool import (
    ToolArgumentErrorResult,
    check_is_custom_tool,
    get_tool_name,
)


class TestToolUtilities:
    def test_check_is_custom_tool_with_callable(self):
        """Test that callable objects are identified as custom tools."""

        async def my_tool():
            pass

        assert check_is_custom_tool(my_tool) is True

    def test_get_tool_name_from_callable(self):
        """Test getting tool name from a callable."""

        async def my_custom_tool():
            pass

        assert get_tool_name(my_custom_tool) == "my_custom_tool"


class TestToolArgumentErrorResult:
    def test_error_result_creation(self):
        """Test creating a tool argument error result."""
        result = ToolArgumentErrorResult(error="Invalid argument")
        assert result.error == "Invalid argument"
        assert result.success is False

    def test_error_result_json_serialization(self):
        """Test that error result can be serialized to JSON."""
        result = ToolArgumentErrorResult(error="Test error")
        json_str = result.model_dump_json()
        assert "Test error" in json_str
        assert "false" in json_str.lower()



import asyncio
from types import NoneType
from typing import Any

import temporalio
import temporalio.workflow
from pydantic import BaseModel

from mistralai_workflows.core.activity import get_wrapped_activity
from mistralai_workflows.core.definition.validation.validator import (
    check_is_valid_type,
    get_function_signature_type_hints,
)
from mistralai_workflows.core.execution.concurrency.types import ChainExecutorParams, WorkflowParams, WorkflowResults
from mistralai_workflows.core.execution.concurrency.utils import dict_to_workflow_results, workflow_results_to_dict


async def execute_chain_activities(
    params: ChainExecutorParams,
) -> WorkflowParams | WorkflowResults:
    """Execute activities in parallel for a chain of items.

    This executor processes items sequentially by fetching the next item
    from the previous one. Use this for token-based pagination (S3, DynamoDB)
    where each response contains the key to get the next batch.

    Args:
        params: Chain executor parameters containing the chain configuration

    Returns:
        WorkflowResults if processing is complete, or WorkflowParams for continuation
    """
    activity = get_wrapped_activity(params.activity_name)
    get_item_from_prev_item_activity = get_wrapped_activity(params.get_item_from_prev_item_activity_name)

    assert activity is not None
    assert get_item_from_prev_item_activity is not None

    user_params_dict, return_type = get_function_signature_type_hints(activity, is_method=False)
    param_type = next(iter(user_params_dict.values())) if user_params_dict else None
    assert param_type is not None and check_is_valid_type(param_type, BaseModel)

    current_item = params.prev_item
    if current_item and not isinstance(current_item, param_type):
        current_item = param_type.model_validate(current_item)

    results_as_dict = workflow_results_to_dict(params.prev_results or WorkflowResults(values=[]))
    idx = params.prev_idx + 1
    done = False

    async def run_activity(idx: int, item: Any) -> None:
        result = await activity(item)
        if return_type is not NoneType:
            results_as_dict[idx] = result

    async with asyncio.TaskGroup() as tg:
        while True:
            current_item = await get_item_from_prev_item_activity(current_item)

            if current_item is None:
                done = True
                break

            tg.create_task(run_activity(idx, current_item))

            if temporalio.workflow.in_workflow() and temporalio.workflow.info().is_continue_as_new_suggested():
                break

            idx += 1

    if done:
        # Stream is exhausted
        return dict_to_workflow_results(results_as_dict)
    else:
        # Need to continue streaming
        return WorkflowParams(
            params=ChainExecutorParams(
                activity_name=params.activity_name,
                get_item_from_prev_item_activity_name=params.get_item_from_prev_item_activity_name,
                prev_idx=idx,
                prev_item=current_item,
                prev_results=dict_to_workflow_results(results_as_dict),
            )
        )



from typing import Any

import pytest

from mistralai_workflows import get_workflow_definition

from .fixtures_activity_extends import (
    ActivityExtendsWorkflow,
    V1Config,
    WorkflowParams,
    process_base,
    process_v1,
    process_v2,
)
from .utils import create_test_worker


class TestActivityExtends:
    @pytest.mark.asyncio
    async def test_extends_decorator_with_temporal_workflow(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[ActivityExtendsWorkflow],
            activities=[process_base, process_v1, process_v2],
        ):
            workflow_def = get_workflow_definition(ActivityExtendsWorkflow)
            assert workflow_def is not None

            params = WorkflowParams(config=V1Config(), data="test")

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                params.model_dump(),
                id="test-activity-extends",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            assert result["result"] == "v1: test"



import datetime
import inspect
from contextvars import ContextVar
from functools import partial, singledispatch, wraps
from typing import Any, Callable, Dict, List, TypeVar, cast

import structlog
import temporalio
import temporalio.activity
import temporalio.common
import temporalio.workflow
import tenacity

from mistralai_workflows.core.config.config import config
from mistralai_workflows.core.definition.validation.validator import (
    get_function_signature_type_hints,
    raise_if_function_has_invalid_return_type,
    raise_if_function_has_invalid_signature,
    raise_if_function_has_invalid_usage,
)
from mistralai_workflows.core.dependencies.dependency_injector import DependencyInjector
from mistralai_workflows.core.execution.local_activity import is_local_activity_enabled
from mistralai_workflows.core.execution.sticky_session.sticky_worker_session import (
    set_activity_as_sticky_to_worker,
)
from mistralai_workflows.core.rate_limiting.rate_limit import (
    RateLimit,
    get_rate_limit,
    local_wait_rate_limit,
    set_rate_limit,
)
from mistralai_workflows.core.utils.contextvars import reset_contextvar, unwrap_contextual_result
from mistralai_workflows.exceptions import ErrorCode, WorkflowsException

logger = structlog.get_logger(__name__)


_temporal_activities: List[Callable] = []
_activities_by_name: Dict[str, Callable] = {}

context_var_task_queue: ContextVar[str | None] = ContextVar("task_queue", default=None)
_context_var_is_inside_activity: ContextVar[bool] = ContextVar("is_inside_activity", default=False)

T = TypeVar("T", bound=Callable[..., Any] | type)


def activity(
    start_to_close_timeout: datetime.timedelta = datetime.timedelta(minutes=5),
    retry_policy_max_attempts: int | None = None,
    retry_policy_backoff_coefficient: float | None = None,
    display_name: str | None = None,
    name: str | None = None,
    sticky_to_worker: bool = False,
    rate_limit: RateLimit | None = None,
    _skip_registering: bool = False,
    activity_name: str | None = None,
    _extends: Callable | None = None,
) -> Callable[[T], T]:
    """
    Decorator for defining activities with configuration options.

    This decorator transforms a regular Python function into an activity with
    various configuration options for timeouts, retries, and worker scoping.

    Args:
        start_to_close_timeout: Maximum time the activity is allowed to execute.
        retry_policy_max_attempts: Maximum number of retry attempts.
        retry_policy_backoff_coefficient: Backoff coefficient for retry delays.
        display_name: Optional display name for the activity when recording events.
        name: Optional name override for the activity. Defaults to function name.
        sticky_to_worker: Whether the activity should be sticky to a worker.
                          See `workflow_sdk.worker.sticky_worker_session` for more details.
        rate_limit: Set a rate limit for the activity across all workers.
                    See `workflow_sdk.worker.rate_limit.RateLimit` for more details.
        activity_name: DEPRECATED. Use 'name' instead.
        _extends: DEPRECATED. Optional activity to extend functionality from.
                  Extended activity must have 2 parameters (`config` and `params`).
                  The activity will be registered as a singledispatch function and be run
                  based on the type of the config parameter provided.
                  See `examples/activity_extends.py` for more details.

    Internal only args:
        _skip_registering: Whether to skip automatic registration of this activity.
                           When set to `True`, the activity will not be returned by `get_all_temporal_activities()`.

    Returns:
        A decorator function that transforms the target function into an activity.
    """
    actual_name = name if name is not None else activity_name

    if activity_name is not None:
        logger.warning(
            "'activity_name' parameter is deprecated, use 'name' instead",
            activity_name=activity_name,
        )

    if _extends is not None:
        logger.warning(
            "'_extends' parameter is deprecated and will be removed in a future version",
            _extends=_extends.__name__ if hasattr(_extends, "__name__") else str(_extends),
        )

    internal_event = display_name is None

    def decorator(target: T) -> T:
        # Only support module-level functions
        if not inspect.isfunction(target):
            raise WorkflowsException(
                code=ErrorCode.ACTIVITY_NOT_MODULE_LEVEL,
                message=f"@workflows.activity only supports module-level functions, got {type(target)}",
            )
        else:
            return cast(
                T,
                _decorate_activity(
                    target,
                    start_to_close_timeout,
                    retry_policy_max_attempts,
                    retry_policy_backoff_coefficient,
                    execution_name=display_name or target.__name__,
                    is_internal_activity=internal_event,
                    activity_name=actual_name,
                    extends=_extends,
                    sticky_to_worker=sticky_to_worker,
                    rate_limit=rate_limit,
                    _skip_registering=_skip_registering,
                ),
            )

    return decorator


def _decorate_activity(
    func: Callable,
    start_to_close_timeout: datetime.timedelta,
    retry_policy_max_attempts: int | None,
    retry_policy_backoff_coefficient: float | None,
    execution_name: str,
    is_internal_activity: bool,
    activity_name: str | None,
    extends: Callable | None,
    sticky_to_worker: bool,
    rate_limit: RateLimit | None,
    _skip_registering: bool,
) -> Callable:
    """
    See `activity` decorator for documentation.
    """
    if activity_name is not None:
        func.__name__ = activity_name
    if retry_policy_max_attempts is None:
        retry_policy_max_attempts = config.worker.retry_policy_max_attempts
    if retry_policy_backoff_coefficient is None:
        retry_policy_backoff_coefficient = config.worker.retry_policy_backoff_coefficient

    # Validate signature BEFORE dependency injection modifies it
    raise_if_function_has_invalid_signature(func, is_method=False)

    user_params_dict, _ = get_function_signature_type_hints(func, is_method=False)

    original_func = func

    dependency_injector = DependencyInjector.get_singleton_instance()
    func = dependency_injector.auto_resolve_dependencies(func)
    # TODO: remove when remove workflow_sdk/workflows/worker/streaming/streaming_context.py
    # func = auto_stream_activity_status(
    #     func, activity_name=execution_name, retry_policy_max_attempts=retry_policy_max_attempts
    # )
    func = temporalio.activity.defn(func)

    if sticky_to_worker:
        set_activity_as_sticky_to_worker(func)

    if rate_limit is not None:
        set_rate_limit(func, rate_limit)
    elif get_rate_limit(func) is not None:
        raise ValueError(
            "Cannot use `@rate_limit` with `@activity`, please use the parameter `rate_limit` instead in `@activity`."  # noqa: E501
        )

    if not _skip_registering:
        # Register activity for temporal worker
        _temporal_activities.append(func)

    @wraps(func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:
        """This wrapper is used to execute an activity from a Workflow."""
        is_inside_activity = _context_var_is_inside_activity.get(False)
        token = _context_var_is_inside_activity.set(True)

        try:
            raise_if_function_has_invalid_usage(original_func, args, kwargs, is_method=False)

            if temporalio.workflow.in_workflow():
                # Determine task queue for activity scheduling
                if sticky_to_worker:
                    task_queue = context_var_task_queue.get()
                    if task_queue is None:
                        raise WorkflowsException(
                            code=ErrorCode.STICKY_WORKER_SESSION_MISSING,
                            message=f"Activity '{func.__name__}' is sticky to worker but no task queue is set. "
                            "Please call this activity inside `run_sticky_worker_session` context manager.",
                        )
                elif rate_limit is not None:
                    rate_limit_config = get_rate_limit(func)
                    if rate_limit_config is None:
                        raise ValueError(f"Activity '{func.__name__}' has rate limit but no rate limit config is set.")
                    task_queue = rate_limit_config.task_queue
                else:
                    task_queue = config.temporal.task_queue
                logger.debug(
                    "Executing activity inside workflow",
                    activity_name=func.__name__,
                    task_queue=task_queue,
                )
                if is_local_activity_enabled():
                    if rate_limit is not None:
                        logger.warning(
                            "Rate limiting is not supported for local activities",
                            activity_name=func.__name__,
                            rate_limit_config=rate_limit.model_dump()
                            if hasattr(rate_limit, "model_dump")
                            else str(rate_limit),
                        )
                    temporal_execute_activity = temporalio.workflow.execute_local_activity
                else:
                    temporal_execute_activity = partial(temporalio.workflow.execute_activity, task_queue=task_queue)

                result = await temporal_execute_activity(
                    func,
                    args=args,
                    start_to_close_timeout=start_to_close_timeout,
                    retry_policy=temporalio.common.RetryPolicy(
                        maximum_attempts=retry_policy_max_attempts,
                        backoff_coefficient=retry_policy_backoff_coefficient,
                    ),
                )
            else:
                if rate_limit is not None:
                    rate_limit_config = get_rate_limit(func)
                    if rate_limit_config is None:
                        raise ValueError(f"Activity '{func.__name__}' has rate limit but no rate limit config is set.")
                    await local_wait_rate_limit(rate_limit_config)

                local_func = (
                    func
                    if temporalio.activity.in_activity() or is_inside_activity
                    else tenacity.retry(
                        retry=tenacity.retry_if_exception_type(Exception),
                        stop=tenacity.stop_after_attempt(retry_policy_max_attempts),
                        wait=tenacity.wait_exponential(multiplier=retry_policy_backoff_coefficient),
                        reraise=True,
                    )(func)
                )

                if DependencyInjector.is_inside_dependencies_context():
                    logger.debug(
                        "Executing activity outside workflow (inside dependencies context)",
                        activity_name=func.__name__,
                    )
                    result = await local_func(*args, **kwargs)
                else:
                    logger.debug(
                        "Executing activity outside workflow (outside dependencies context)",
                        activity_name=func.__name__,
                    )
                    dependency_injector = DependencyInjector.get_singleton_instance()
                    async with dependency_injector.with_dependencies():
                        result = await local_func(*args, **kwargs)

            _, result = unwrap_contextual_result(result)
            raise_if_function_has_invalid_return_type(original_func, result, is_method=False)

        finally:
            reset_contextvar(_context_var_is_inside_activity, token)

        return result

    if hasattr(func, "__temporal_activity_definition"):
        wrapper.__temporal_activity_definition = func.__temporal_activity_definition  # type: ignore[attr-defined]

    if extends is not None:
        if not hasattr(extends, "register"):
            raise ValueError(
                f"activity '{extends.__name__}' must be decorated with @workflows.activity and contains 2 parameters (`config` and `params`) to be used inside extends."  # noqa: E501
            )

        wrapper = extends.register(wrapper)  # type: ignore[attr-defined]
    elif len(user_params_dict) == 2:
        param_names = list(user_params_dict.keys())
        if param_names == ["config", "params"]:
            wrapper = singledispatch(wrapper)  # type: ignore[assignment]

    if not _skip_registering:
        # Register activity for internal usage
        _activities_by_name[wrapper.__name__] = wrapper

    return cast(Callable, wrapper)


def get_all_temporal_activities() -> List[Callable]:
    """
    Get a list of all registered Temporal activities.

    Returns:
        List[Callable]: A list of all activity functions registered with Temporal.
    """
    return _temporal_activities.copy()  # return a copy to avoid modifying the original list


def get_wrapped_activity(activity_name: str) -> Callable | None:
    """
    Get a wrapped activity function by name.

    Args:
        activity_name: The name of the activity to retrieve.

    Returns:
        Callable | None: The wrapped activity function if found, None otherwise.
    """
    return _activities_by_name.get(activity_name)


def check_is_activity(func: Callable) -> bool:
    """
    Check if a function is a registered Temporal activity.

    Args:
        func: The function to check.

    Returns:
        bool: True if the function is a registered activity, False otherwise.
    """
    return func in _activities_by_name.values()



from typing import Any

import pytest

from mistralai_workflows import Depends, activity, get_workflow_definition, workflow
from mistralai_workflows.core.execution.local_activity import run_activities_locally

from .utils import create_test_worker


class DatabaseService:
    def __init__(self, connection_string: str = "test-db"):
        self.connection_string = connection_string
        self.queries: list[str] = []

    async def query(self, sql: str) -> str:
        self.queries.append(sql)
        return f"DB[{self.connection_string}]: {sql}"


class LogService:
    def __init__(self):
        self.logs: list[str] = []
        self.call_count = 0

    def log(self, message: str) -> None:
        self.logs.append(message)
        self.call_count += 1


def get_database() -> DatabaseService:
    return DatabaseService("injected-db-connection")


def get_logger() -> LogService:
    logger = LogService()
    logger.was_injected = True  # type: ignore[attr-defined]
    return logger


@activity()
async def process_with_db(
    text: str,
    db: DatabaseService = Depends(get_database),
) -> str:
    result = await db.query(f"SELECT * FROM messages WHERE text='{text}'")
    return result


@activity()
async def process_with_multiple_deps(
    text: str,
    db: DatabaseService = Depends(get_database),
    logger: LogService = Depends(get_logger),
) -> str:
    assert hasattr(logger, "was_injected"), "Logger was not properly injected"

    logger.log(f"Processing: {text}")
    result = await db.query(f"INSERT INTO messages VALUES ('{text}')")
    logger.log(f"Query executed: {result}")

    return f"{result} [logged {logger.call_count} times]"


@workflow.define(name="test-workflow-with-dependency-injection")
class WorkflowWithDependencyInjection:
    @workflow.entrypoint
    async def run(self, text: str) -> str:
        result = await process_with_db(text)
        return result


@workflow.define(name="test-workflow-with-multiple-deps")
class WorkflowWithMultipleDependencies:
    @workflow.entrypoint
    async def run(self, text: str) -> str:
        result = await process_with_multiple_deps(text)
        return result


@workflow.define(name="test-workflow-with-local-deps")
class WorkflowWithLocalDependencies:
    @workflow.entrypoint
    async def run(self, text: str) -> str:
        with run_activities_locally():
            result = await process_with_multiple_deps(text)
        return result


class TestDependencyInjection:
    @pytest.mark.asyncio
    async def test_activity_with_injected_dependency(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowWithDependencyInjection],
            activities=[process_with_db],
        ):
            workflow_def = get_workflow_definition(WorkflowWithDependencyInjection)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "test message"},
                id="test-di-workflow",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            assert isinstance(result, dict)
            assert "DB[injected-db-connection]" in result["result"]
            assert "SELECT * FROM messages" in result["result"]
            assert "test message" in result["result"]

    @pytest.mark.asyncio
    async def test_activity_with_multiple_dependencies(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowWithMultipleDependencies],
            activities=[process_with_multiple_deps],
        ):
            workflow_def = get_workflow_definition(WorkflowWithMultipleDependencies)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "multi-dep message"},
                id="test-multi-di-workflow",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            assert isinstance(result, dict)
            assert "DB[injected-db-connection]" in result["result"]
            assert "INSERT INTO messages" in result["result"]
            assert "[logged 2 times]" in result["result"]

    @pytest.mark.asyncio
    async def test_dependency_isolation_between_workflows(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowWithDependencyInjection],
            activities=[process_with_db],
        ):
            workflow_def = get_workflow_definition(WorkflowWithDependencyInjection)
            assert workflow_def is not None

            handle1 = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "first"},
                id="test-di-isolation-1",
                task_queue="test-task-queue",
            )
            result1 = await handle1.result()

            handle2 = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "second"},
                id="test-di-isolation-2",
                task_queue="test-task-queue",
            )
            result2 = await handle2.result()

            assert "first" in result1["result"]
            assert "second" in result2["result"]

    @pytest.mark.asyncio
    async def test_dependencies_not_in_input_schema(self, temporal_env: Any) -> None:
        workflow_def = get_workflow_definition(WorkflowWithDependencyInjection)

        assert workflow_def.input_schema is not None
        properties = workflow_def.input_schema.get("properties", {})

        assert "params" in properties or "text" in properties
        assert "db" not in properties
        assert "logger" not in properties

    @pytest.mark.asyncio
    async def test_dependency_injection_works_with_local_activities(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[WorkflowWithLocalDependencies],
            activities=[process_with_multiple_deps],
        ):
            workflow_def = get_workflow_definition(WorkflowWithLocalDependencies)
            assert workflow_def is not None

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                {"text": "local activity test"},
                id="test-local-di-workflow",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            assert isinstance(result, dict)
            assert "DB[injected-db-connection]" in result["result"]
            assert "INSERT INTO messages" in result["result"]
            assert "[logged 2 times]" in result["result"]



from typing import Any

import pytest
import temporalio.activity
from pydantic import BaseModel

from mistralai_workflows import activity, get_workflow_definition, workflow
from mistralai_workflows.core.execution.local_activity import run_activities_locally

from .utils import create_test_worker


class SimpleParams(BaseModel):
    value: str


class ActivityResult(BaseModel):
    is_local_activity: bool


def check_is_activity_running_locally():
    info = temporalio.activity.info()
    return info.is_local


@activity()
async def activity_1(params: SimpleParams) -> ActivityResult:
    return ActivityResult(is_local_activity=check_is_activity_running_locally())


@activity()
async def activity_2(params: SimpleParams) -> ActivityResult:
    return ActivityResult(is_local_activity=check_is_activity_running_locally())


@activity()
async def activity_3(params: SimpleParams) -> ActivityResult:
    return ActivityResult(is_local_activity=check_is_activity_running_locally())


@activity()
async def activity_4(params: SimpleParams) -> ActivityResult:
    return ActivityResult(is_local_activity=check_is_activity_running_locally())


@workflow.define(name="test-full-remote-workflow")
class FullRemoteWorkflow:
    @workflow.entrypoint
    async def run(self, params: SimpleParams) -> list[ActivityResult]:
        results = [
            await activity_1(params),
            await activity_2(params),
            await activity_3(params),
        ]
        return results


@workflow.define(name="test-full-local-workflow")
class FullLocalWorkflow:
    @workflow.entrypoint
    async def run(self, params: SimpleParams) -> list[ActivityResult]:
        with run_activities_locally():
            results = [
                await activity_1(params),
                await activity_2(params),
                await activity_3(params),
            ]
        return results


@workflow.define(name="test-mixed-execution-workflow")
class MixedExecutionWorkflow:
    @workflow.entrypoint
    async def run(self, params: SimpleParams) -> list[ActivityResult]:
        results = [
            await activity_1(params)  # Remote
        ]

        with run_activities_locally():
            results.extend(
                [
                    await activity_2(params),  # Local
                    await activity_3(params),  # Local
                ]
            )

        results.append(await activity_1(params))  # Remote
        return results


class TestLocalActivities:
    @pytest.mark.asyncio
    async def test_full_remote_execution(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[FullRemoteWorkflow],
            activities=[activity_1, activity_2, activity_3],
        ):
            workflow_def = get_workflow_definition(FullRemoteWorkflow)
            assert workflow_def is not None

            params = SimpleParams(value="test-remote")

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                params.model_dump(),
                id="test-full-remote",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            results = result["result"]
            assert len(results) == 3
            for activity_result in results:
                assert activity_result["is_local_activity"] is False

    @pytest.mark.asyncio
    async def test_full_local_execution(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[FullLocalWorkflow],
            activities=[activity_1, activity_2, activity_3],
        ):
            workflow_def = get_workflow_definition(FullLocalWorkflow)
            assert workflow_def is not None

            params = SimpleParams(value="test-local")

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                params.model_dump(),
                id="test-full-local",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            results = result["result"]
            assert len(results) == 3
            for activity_result in results:
                assert activity_result["is_local_activity"] is True

    @pytest.mark.asyncio
    async def test_mixed_execution(self, temporal_env: Any) -> None:
        async with create_test_worker(
            temporal_env,
            workflows=[MixedExecutionWorkflow],
            activities=[activity_1, activity_2, activity_3],
        ):
            workflow_def = get_workflow_definition(MixedExecutionWorkflow)
            assert workflow_def is not None

            params = SimpleParams(value="test-mixed")

            handle = await temporal_env.client.start_workflow(
                workflow_def.name,
                params.model_dump(),
                id="test-mixed-execution",
                task_queue="test-task-queue",
            )

            result = await handle.result()

            results = result["result"]
            assert len(results) == 4

            expected_is_local = [False, True, True, False]
            for i, (activity_result, expected) in enumerate(zip(results, expected_is_local)):
                assert activity_result["is_local_activity"] == expected, (
                    f"Activity {i} local={activity_result['is_local_activity']}, expected={expected}"
                )



import importlib
import pkgutil
from importlib.metadata import packages_distributions, version
from typing import NamedTuple, cast

PLUGIN_NAMESPACE = cast(str, __package__)  # cast so it's not typed as optional str
SELF = __name__.split(".")[-1]


class ContributionInfo(NamedTuple):
    name: str
    ispkg: bool
    dist_name: str | None
    dist_version: str | None


def list_plugins() -> list[ContributionInfo]:
    """List names and versions of all installed plugins."""
    pkg_to_dist = packages_distributions()
    results: list[ContributionInfo] = []
    for module_info in pkgutil.iter_modules(importlib.import_module(PLUGIN_NAMESPACE).__path__):
        if module_info.name != SELF:
            import_name = f"{PLUGIN_NAMESPACE}.{module_info.name}"
            dist_names = pkg_to_dist.get(import_name, [])
            (dist_name, dist_version) = (dist_names[0], version(dist_names[0])) if dist_names else (None, None)
            results.append(ContributionInfo(module_info.name, module_info.ispkg, dist_name, dist_version))
    return results



"""
OLD Insurance Claims Workflow Example

This file contains the OLD implementation using workflows.streaming.stream() API.
It is kept for reference to show the old way of coding things.

The new approach uses the Task API from workflows.worker.task instead.

DO NOT USE THIS CODE - it relies on deprecated streaming APIs.

---

OLD CODE (kept for reference):

```python
# 10,153 characters
# we keep it for later comparison with the new implementation
import asyncio
import os
from textwrap import dedent
from typing import Any, List, Literal

import mistralai

# Import actual mistralai types for proper type checking
from mistralai import SystemMessage, UserMessage
from pydantic import BaseModel, Field

import mistralai_workflows as workflows


class LeChatPayloadWorking(BaseModel):
    type: Literal["working"] = "working"
    name: str
    results: str | None = None


class LeChatPayloadAssistantMessage(mistralai.AssistantMessage):
    type: Literal["assistant_message"] = "assistant_message"


class LeChatPayloadHumanFeedback(BaseModel):
    type: Literal["human_feedback"] = "human_feedback"
    input_schema: dict
    input: Any | None = None


class WaitUserInputUpdateParams(BaseModel):
    custom_task_id: str
    input: Any


class WaitUserInputUpdateResult(BaseModel):
    error: str | None = None


class UserInputRequest(BaseModel):
    custom_task_id: str
    input_schema: type[BaseModel]
    received: bool = False
    input: Any | None = None


class ClaimType(BaseModel):
    claim_type: Literal["medical", "auto_accident", "other"] = "other"


class ClaimStatus(BaseModel):
    status: Literal["approved", "rejected", "pending"] = "pending"


class ClaimResponseSignalParams(BaseModel):
    is_approved: bool = Field(description="Whether the claim is approved or rejected", default=False)


class WorkflowParams(BaseModel):
    attachement_urls: List[str]
    email_content: str


class WorkflowOutput(BaseModel):
    final_email: str


@workflows.workflow.define(
    name="insurance_claims_workflow",
    workflow_description="Workflow to process insurance claims",
)
class InsuranceClaimsWorkflow:
    def __init__(self) -> None:
        self._user_input_requests: dict[str, UserInputRequest] = {}

    async def _wait_for_input(self, input_schema: type[BaseModel]) -> Any:
        \"\"\"Helper method to wait for user input with the old-style pattern.\"\"\"
        with workflows.streaming.stream("human_feedback") as feedback_stream:
            feedback_stream.publish(LeChatPayloadHumanFeedback(input_schema=input_schema.model_json_schema()))
            self._user_input_requests[feedback_stream.custom_task_id] = UserInputRequest(
                custom_task_id=feedback_stream.custom_task_id,
                input_schema=input_schema,
            )
            await workflows.workflow.wait_condition(
                lambda task_id=feedback_stream.custom_task_id: bool(
                    self._user_input_requests[task_id].received
                )  # type: ignore[misc]
            )
            user_input = input_schema.model_validate(
                self._user_input_requests[feedback_stream.custom_task_id].input
            )
            feedback_stream.publish(
                LeChatPayloadHumanFeedback(
                    input_schema=input_schema.model_json_schema(),
                    input=user_input,
                )
            )
            return user_input

    @workflows.workflow.entrypoint
    async def run(self, params: WorkflowParams) -> WorkflowOutput:
        # Extract attachments content (OCR)
        attachements_content: dict[str, str] = {}
        if params.attachement_urls:
            with workflows.streaming.stream("working") as stream:
                stream.publish(LeChatPayloadWorking(name="OCR attachments"))
                ocr_tasks = [llm_ocr_from_url(MistralOCRParams(url=url)) for url in params.attachement_urls]
                responses = await asyncio.gather(*ocr_tasks)
                stream.publish(
                    LeChatPayloadWorking(name="OCR attachments", results=f"Processed {len(responses)} documents")
                )
            attachements_content = {
                url: "\\n\\n".join([page.markdown for page in response.pages])
                for url, response in zip(params.attachement_urls, responses, strict=True)
            }

        # Build full email with attachments
        documents = "\\n\\n".join(
            f"--- Document: {url}\\n\\n{markdown}" for url, markdown in attachements_content.items()
        )
        full_email = dedent(
            f\"\"\"
            <received_email>
            {params.email_content}
            </received_email>

            <received_attachments>
            {documents}
            </received_attachments>
            \"\"\"
        )

        # Classify claim type
        with workflows.streaming.stream("working") as stream:
            stream.publish(LeChatPayloadWorking(name="Categorizing claim"))
            try:
                claim_type = await llm_classify_claim_type(LLMClassifyClaimTypeParams(full_email=full_email))
            except Exception:
                claim_type = await self._wait_for_input(ClaimType)
            stream.publish(LeChatPayloadWorking(name="Categorizing claim", results=claim_type.claim_type))

        # Predict claim status
        with workflows.streaming.stream("working") as stream:
            stream.publish(LeChatPayloadWorking(name="Predicting claim status"))
            if claim_type.claim_type == "auto_accident":
                # Auto-approve auto accident claims
                claim_status = ClaimStatus(status="approved")
            elif claim_type.claim_type == "medical":
                # Ask for human feedback for medical claims
                claim_status = await self._wait_for_input(ClaimStatus)
            elif claim_type.claim_type == "other":
                # Reject other claims
                claim_status = ClaimStatus(status="rejected")
            else:
                raise ValueError("Invalid claim category")
            stream.publish(LeChatPayloadWorking(name="Predicting claim status", results=claim_status.status))

        # Generate final email
        final_email = await llm_generate_final_email(
            GenerateEmailParams(full_email=full_email, claim_status=claim_status)
        )
        return WorkflowOutput(final_email=final_email.content)

    @workflows.workflow.update(name="human_feedback")
    async def human_feedback(self, message: WaitUserInputUpdateParams) -> WaitUserInputUpdateResult:
        user_input_request = self._user_input_requests.get(message.custom_task_id)
        if not user_input_request:
            return WaitUserInputUpdateResult(error=f"Custom task {message.custom_task_id} not found")

        try:
            user_input_request.input_schema.model_validate(message.input)
        except Exception as e:
            return WaitUserInputUpdateResult(error=f"Invalid input: {e}")

        user_input_request.input = message.input
        user_input_request.received = True
        return WaitUserInputUpdateResult()


def get_mistral_client() -> mistralai.Mistral:
    return mistralai.Mistral(api_key=os.getenv("MISTRAL_API_KEY"))


class LLMClassifyClaimTypeParams(BaseModel):
    full_email: str


@workflows.activity()
async def llm_classify_claim_type(
    params: LLMClassifyClaimTypeParams,
    mistral_client: mistralai.Mistral = workflows.Depends(get_mistral_client),
) -> ClaimType:
    with workflows.streaming.stream("assistant_message") as stream:
        stream.publish(LeChatPayloadAssistantMessage(content="Classifying claim type..."))
        result = await mistral_client.chat.parse_async(
            model="mistral-medium-latest",
            messages=[
                mistralai.SystemMessage(content="Given the following claim, extract the category of the claim."),
                mistralai.UserMessage(content=params.full_email),
            ],
            response_format=ClaimType,
        )
        if not result.choices or not result.choices[0].message or not result.choices[0].message.parsed:
            stream.publish(LeChatPayloadAssistantMessage(content="Failed to classify claim type"))
            raise ValueError("No response from Mistral")
        else:
            parsed = result.choices[0].message.parsed
            stream.publish(LeChatPayloadAssistantMessage(content=parsed.model_dump_json(indent=2)))
            return parsed


class GenerateEmailParams(BaseModel):
    full_email: str
    claim_status: ClaimStatus


class GenerateEmailResult(BaseModel):
    content: str


@workflows.activity()
async def llm_generate_final_email(
    params: GenerateEmailParams,
    mistral_client: mistralai.Mistral = workflows.Depends(get_mistral_client),
) -> GenerateEmailResult:
    with workflows.streaming.stream("assistant_message") as stream:
        assistant_message_content = ""
        stream.publish(LeChatPayloadAssistantMessage(content=assistant_message_content))
        from mistralai import AssistantMessage, ToolMessage

        messages: list[SystemMessage | UserMessage | AssistantMessage | ToolMessage] = [
            SystemMessage(content="Generate a short email to the customer with the claim conclusion."),
            UserMessage(
                content=dedent(
                    f\"\"\"
                    The claim has been {params.claim_status.status}.
                    Sign the email with 'Best regards, Insurance Company'.
                    Here is the original claim information:

                    {params.full_email}
                    \"\"\"
                )
            ),
        ]
        mistral_stream = await mistral_client.chat.stream_async(
            model="mistral-medium-2508",
            messages=messages,
        )
        async for chunk in mistral_stream.generator:
            if chunk.data.choices[0].delta.content:
                assert isinstance(chunk.data.choices[0].delta.content, str), "Non string content is not supported"
                assistant_message_content += chunk.data.choices[0].delta.content
                stream.publish(LeChatPayloadAssistantMessage(content=assistant_message_content))

        return GenerateEmailResult(content=assistant_message_content)


class MistralOCRParams(BaseModel):
    mistral_model: str = "mistral-ocr-latest"
    url: str


@workflows.activity()
async def llm_ocr_from_url(
    params: MistralOCRParams,
    mistral_client: mistralai.Mistral = workflows.Depends(get_mistral_client),
) -> mistralai.OCRResponse:
    \"\"\"Perform OCR on a document using Mistral's OCR API.\"\"\"
    document = mistralai.DocumentURLChunk(document_url=params.url)
    response = await mistral_client.ocr.process_async(
        model=params.mistral_model,
        document=document,
        include_image_base64=True,
    )
    return response


if __name__ == "__main__":
    asyncio.run(workflows.run_worker([InsuranceClaimsWorkflow]))
```
"""



import uuid
from enum import StrEnum
from typing import Any, Dict, List, Literal

from pydantic import BaseModel, Field

from .handlers import QueryDefinition, SignalDefinition, UpdateDefinition
from .schedule import ScheduleDefinition


class WorkflowCodeDefinition(BaseModel):
    input_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Input schema of the workflow's run method",
        json_schema_extra={"additionalProperties": True},
    )  # note change to Dict[str, Any] which is closer to real Json type.
    output_schema: Dict[str, Any] | None = Field(
        default=None,
        description="Output schema of the workflow's run method",
        json_schema_extra={"additionalProperties": True},
    )
    signals: List[SignalDefinition] = Field(default_factory=list, description="Signal handlers defined by the workflow")
    queries: List[QueryDefinition] = Field(default_factory=list, description="Query handlers defined by the workflow")
    updates: List[UpdateDefinition] = Field(default_factory=list, description="Update handlers defined by the workflow")


class WorkflowSpec(WorkflowCodeDefinition):
    name: str = Field(description="Name of the workflow")
    display_name: str | None = Field(default=None, description="Display name of the workflow")
    description: str | None = Field(default=None, description="Description of the workflow")
    schedules: List[ScheduleDefinition] = Field(default_factory=list, description="Schedules defined by the workflow")


class WorkflowSpecWithTaskQueue(WorkflowSpec):
    task_queue: str = Field(description="Task queue name for the workflow")


class WorkflowType(StrEnum):
    CODE = "code"
    # DSL = "dsl"


class Workflow(BaseModel):
    id: uuid.UUID = Field(description="Unique identifier of the workflow")
    name: str = Field(description="Name of the workflow")
    display_name: str = Field(description="Display name of the workflow")
    type: WorkflowType = Field(description="Type of the workflow")
    description: str | None = Field(default=None, description="Description of the workflow")
    customer_id: uuid.UUID = Field(description="Customer ID of the workflow")
    workspace_id: uuid.UUID = Field(description="Workspace ID of the workflow")
    shared_namespace: str | None = Field(
        default=None, description="Reserved namespace for shared workflows (e.g., 'shared:my-shared-workflow')"
    )
    available_in_chat_assistant: bool = Field(description="Whether the workflow is available in chat assistant")


class WorkflowVersion(BaseModel):
    id: uuid.UUID = Field(description="Unique identifier of the workflow version")
    task_queue: str = Field(description="Project name of the workflow")
    definition: WorkflowCodeDefinition
    workflow_id: uuid.UUID = Field(description="Workflow ID of the workflow")
    workflow: Workflow | None = Field(default=None, description="Workflow of the workflow version")
    compatible_with_chat_assistant: bool = Field(
        default=False, description="Whether the workflow is compatible with chat assistant"
    )


class CanvasPayload(BaseModel):
    type: Literal[
        "text/markdown",
        "text/html",
        "image/svg+xml",
        "slides",
        "react",
        "code",
        "mermaid",
    ]
    title: str
    content: str
    language: str | None = None


class CanvasResource(BaseModel):
    uri: str
    mimeType: Literal["application/vnd.mistral.canvas"]
    canvas: CanvasPayload


class ChatAssistantTextChunk(BaseModel):
    type: Literal["text"] = "text"
    text: str


class ChatAssistantResourceChunk(BaseModel):
    type: Literal["resource"] = "resource"
    resource: CanvasResource


ChatAssistantWorkflowOutput = ChatAssistantTextChunk | ChatAssistantResourceChunk



from contextlib import asynccontextmanager
from typing import AsyncGenerator

import structlog

from mistralai_workflows.core.activity import context_var_task_queue
from mistralai_workflows.core.execution.sticky_session.get_sticky_worker_session import get_sticky_worker_session
from mistralai_workflows.core.execution.sticky_session.sticky_worker_session import StickyWorkerSession
from mistralai_workflows.core.utils.contextvars import reset_contextvar

logger = structlog.get_logger(__name__)


@asynccontextmanager
async def run_sticky_worker_session(
    sticky_worker_session: StickyWorkerSession | None = None,
) -> AsyncGenerator[None, None]:
    """
    Route sticky activities to the same worker instance.

    All sticky activities (marked with sticky_to_worker=True) called within this context
    will execute on the same worker, enabling resource reuse and data locality.

    ## Usage Patterns

    Explicit session (can reuse across multiple scopes):
    ```python
    session = await get_sticky_worker_session()
    async with run_sticky_worker_session(session):
        await sticky_activity_1()
        await sticky_activity_2()

    # Later, same worker:
    async with run_sticky_worker_session(session):
        await sticky_activity_3()
    ```

    Implicit session (auto-created):
    ```python
    async with run_sticky_worker_session():
        await sticky_activity_1()
        await sticky_activity_2()
    ```

    ## Nesting Support

    Context managers can be nested. The inner scope temporarily overrides the outer:
    ```python
    async with run_sticky_worker_session(session_a):
        await activity()  # Uses session_a

        async with run_sticky_worker_session(session_b):
            await activity()  # Uses session_b

        await activity()  # Back to session_a
    ```

    Args:
        sticky_worker_session: Worker session to use. If None, automatically creates one
                              by calling get_sticky_worker_session().

    Yields:
        None. The context manager establishes routing state via context variables.
    """
    if sticky_worker_session is None:
        sticky_worker_session = await get_sticky_worker_session()
    token = context_var_task_queue.set(sticky_worker_session.task_queue)
    try:
        yield
    finally:
        reset_contextvar(context_var_task_queue, token)



from mistralai_workflows.core.utils.cache import InMemoryCacheError, SimpleLRUCache, in_memory_cache
from mistralai_workflows.core.utils.contextvars import reset_contextvar, unwrap_contextual_result
from mistralai_workflows.core.utils.id_generator import generate_two_part_id

__all__ = [
    "InMemoryCacheError",
    "SimpleLRUCache",
    "generate_two_part_id",
    "in_memory_cache",
    "reset_contextvar",
    "unwrap_contextual_result",
]



from typing import Any, Protocol, Self, Type


class StatelessTaskProtocol(Protocol):
    """Protocol for tasks without state management."""

    @property
    def id(self) -> str: ...

    @property
    def type(self) -> str: ...

    def __enter__(self) -> Self: ...

    def __exit__(self, exc_type: Type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None: ...


class StatefulTaskProtocol[T](Protocol):
    """Protocol for tasks with state management."""

    @property
    def id(self) -> str: ...

    @property
    def type(self) -> str: ...

    @property
    def state(self) -> T: ...

    def __enter__(self) -> Self: ...

    def __exit__(self, exc_type: Type[BaseException] | None, exc_val: BaseException | None, exc_tb: Any) -> None: ...

    def set_state(self, state: T) -> None: ...

    def update_state(self, updates: dict[str, Any]) -> None: ...



import uuid
from typing import Literal

from pydantic import BaseModel, ConfigDict, Field


class LeChatPayloadWorking(BaseModel):
    model_config = ConfigDict(title="working")

    type: Literal["tool", "thinking"] | str = "tool"
    title: str
    content: str


class LeChatPayloadAssistantMessage(BaseModel):
    model_config = ConfigDict(title="assistant_message")

    role: Literal["assistant"] = "assistant"
    content: str = Field(default="", description="Assistant message content")


class LeChatPayloadUserMessage(BaseModel):
    model_config = ConfigDict(title="user_message")

    role: Literal["user"] = "user"
    content: str = Field(default="", description="User message to continue the conversation")


class LeChatMarkdownOutput(BaseModel):
    mime_type: Literal["text/markdown"] = "text/markdown"
    content: str
    uri: str = Field(default_factory=lambda: f"file://markdown/{uuid.uuid4()}.md")


class LeChatOutput(BaseModel):
    outputs: "list[LeChatPayloadAssistantMessage | LeChatPayloadUserMessage | LeChatMarkdownOutput]" = Field(
        default_factory=list
    )



import asyncio
from typing import List

import pydantic
import structlog
from temporalio import workflow as temporal_workflow

import mistralai_workflows as workflows
from mistralai_workflows.core.logging import Env, LogFormat, LogLevel, setup_logging

logger = structlog.getLogger(__name__)


class AppConfig(pydantic.BaseModel):
    env: Env = Env.DEV
    temporal_server_url: str = "localhost:7233"
    temporal_namespace: str = "default"
    log_format: str = "console"
    log_level: str = "DEBUG"
    app_version: str = "local_test"


class Empty(pydantic.BaseModel):
    pass


class GameWorkflowParams(pydantic.BaseModel):
    initial_player_name: str | None = "AnonPlayer"
    max_riddles: int = 2


class SetNameSignalInput(pydantic.BaseModel):
    new_name: str


class SubmitAnswerSignalInput(pydantic.BaseModel):
    answer: str


class GameWorkflowResult(pydantic.BaseModel):
    final_score: int
    player_name: str
    game_log: List[str]
    message: str


class GameStatusQueryResponse(pydantic.BaseModel):
    player_name: str
    current_riddle: str | None
    riddles_solved: int
    score: int
    game_over: bool
    log_preview: List[str]


class Riddle(pydantic.BaseModel):
    question: str
    correct_answer: str


# --- Activities  ---


@workflows.activity()
async def fetch_next_riddle_activity(solved_count: int) -> Riddle:
    logger.info("activity: fetching riddle", solved_count=solved_count)
    await asyncio.sleep(0.1)
    riddles_db = [
        Riddle(
            question="I have cities, but no houses. I have mountains, but no trees. I have water, \
                but no fish. What am I?",
            correct_answer="map",
        ),
        Riddle(question="What has an eye, but cannot see?", correct_answer="needle"),
        Riddle(question="What is full of holes but still holds water?", correct_answer="sponge"),
    ]
    if solved_count < len(riddles_db):
        return riddles_db[solved_count]
    # Return an empty riddle instead of None to satisfy type checking
    return Riddle(question="No more riddles available", correct_answer="")


# --- Workflow  ---


@workflows.workflow.define(
    name="example-interactive-game-workflow",
    workflow_description="A simple interactive riddle game driven by signals and queries.",
)
class InteractiveGameWorkflow:
    def __init__(self) -> None:
        self.player_name: str = "PlayerUnset"
        self.current_riddle_text: str | None = None
        self.current_riddle_answer: str | None = None
        self.riddles_solved_count: int = 0
        self.score: int = 0
        self.game_over: bool = False
        self.game_log: List[str] = []
        self.max_riddles_to_win: int = 1

        # For wait_for_condition
        self._action_occurred: bool = False

    def _add_log(self, message: str) -> None:
        timestamp = temporal_workflow.now().isoformat()
        entry = f"{timestamp} - {message}"
        self.game_log.append(entry)
        logger.info(f"GameLog[{self.player_name}]: {message}")

    async def _load_new_riddle(self) -> None:
        riddle_data = await fetch_next_riddle_activity(self.riddles_solved_count)

        # Check if the returned riddle is the empty placeholder
        if riddle_data.correct_answer:
            self.current_riddle_text = riddle_data.question
            self.current_riddle_answer = riddle_data.correct_answer
            self._add_log(f"New riddle presented: {self.current_riddle_text}")
        else:
            self._add_log("No more riddles available!")
            self.current_riddle_text = None
            if self.riddles_solved_count >= self.max_riddles_to_win:
                self._add_log(f"Congratulations {self.player_name}! You solved all riddles!")
                self.game_over = True

    @workflows.workflow.entrypoint
    async def run(self, initial_player_name: str | None = "AnonPlayer", max_riddles: int = 2) -> GameWorkflowResult:
        self.player_name = initial_player_name or "Hero"
        self.max_riddles_to_win = max_riddles
        self._add_log(f"Game started for {self.player_name}. Solve {self.max_riddles_to_win} riddle(s) to win.")

        await self._load_new_riddle()

        if self.current_riddle_text is None and self.riddles_solved_count < self.max_riddles_to_win:
            self._add_log("Failed to load initial riddle. Ending game.")
            self.game_over = True

        while not self.game_over:
            self._add_log(
                f"Waiting for player action (current riddle: {'...' if self.current_riddle_text else 'None'})..."
            )
            await workflows.workflow.wait_condition(lambda: self._action_occurred)

            # AND RESET THE FLAG:
            self._action_occurred = False

            if self.game_over:  # A signal might have ended the game
                break

            # Check if we need a new riddle (e.g., after a correct answer processed by a signal)
            # This logic is a bit simplified; signal handlers might directly call _load_new_riddle
            # or set a flag that this loop checks. For now, assume signals update state,
            # and if a riddle was solved, current_riddle_text might be None.
            if self.current_riddle_text is None and not self.game_over:
                self._add_log("Current riddle solved or became unavailable, attempting to load next.")
                await self._load_new_riddle()
                if (
                    self.current_riddle_text is None and not self.game_over
                ):  # Still no riddle and not game over (e.g. no more riddles but not enough solved)
                    self._add_log("No more riddles to solve, but not enough solved to win. Game over.")
                    self.game_over = True

        final_message = f"Game Over! Thanks for playing, {self.player_name}."
        if self.riddles_solved_count >= self.max_riddles_to_win:
            final_message = f"YOU WON, {self.player_name}! You solved {self.riddles_solved_count} riddle(s)!"
        self._add_log(final_message)

        return GameWorkflowResult(
            final_score=self.score, player_name=self.player_name, game_log=self.game_log, message=final_message
        )

    @workflows.workflow.signal(name="set_name", description="Sets the player's name.")
    async def set_name_signal(self, new_name: str) -> None:
        self._add_log(f"Name change requested from '{self.player_name}' to '{new_name}'.")
        self.player_name = new_name
        self._action_occurred = True  # Notify the main loop

    @workflows.workflow.signal(name="submit_answer", description="Submits an answer to the current riddle.")
    async def submit_answer_signal(self, answer: str) -> None:
        if self.game_over:
            self._add_log(f"Answer '{answer}' submitted, but game is already over.")
            self._action_occurred = True
            return

        if not self.current_riddle_text or not self.current_riddle_answer:
            self._add_log(f"Answer '{answer}' submitted, but no active riddle.")
            self._action_occurred = True
            return

        self._add_log(f"Player '{self.player_name}' answered '{answer}' for riddle: '{self.current_riddle_text}'.")
        if answer.lower().strip() == self.current_riddle_answer.lower().strip():
            self.score += 10
            self.riddles_solved_count += 1
            self._add_log(f"Correct! Score: {self.score}. Riddles solved: {self.riddles_solved_count}.")
            self.current_riddle_text = None  # Mark as solved, run loop will try to load next
            if self.riddles_solved_count >= self.max_riddles_to_win:
                self._add_log(f"Victory condition met with {self.riddles_solved_count} riddles solved!")
                self.game_over = True
        else:
            self.score -= 1  # Penalty for wrong answer
            self._add_log(f"Incorrect. Score: {self.score}.")
            # Could add attempts logic here too

        self._action_occurred = True  # Notify the main loop

    @workflows.workflow.signal(name="force_end_game", description="Ends the game immediately.")
    async def force_end_game_signal(self) -> None:  # No input needed
        self._add_log(f"Game force-ended by signal for player {self.player_name}.")
        self.game_over = True
        self._action_occurred = True

    @workflows.workflow.query(name="get_status", description="Gets the current game status.")
    def get_status_query(self) -> GameStatusQueryResponse:
        # Create a preview of the log (handle empty case gracefully)
        log_preview = self.game_log[-5:] if len(self.game_log) > 0 else []

        return GameStatusQueryResponse(
            player_name=self.player_name,
            current_riddle=self.current_riddle_text,
            riddles_solved=self.riddles_solved_count,
            score=self.score,
            game_over=self.game_over,
            log_preview=log_preview,  # Last 5 log entries
        )


# --- Main for local worker (optional, good for quick testing) ---

if __name__ == "__main__":
    app_cfg = AppConfig()
    setup_logging(
        log_format=LogFormat(app_cfg.log_format),
        log_level=LogLevel(app_cfg.log_level),
        app_version=app_cfg.app_version,
    )
    asyncio.run(workflows.run_worker([InteractiveGameWorkflow]))



import time
from enum import StrEnum
from typing import Annotated, Any, Literal

import temporalio.workflow
from pydantic import BaseModel, ConfigDict, Discriminator, Field, Tag


class WorkflowEventType(StrEnum):
    WORKFLOW_EXECUTION_STARTED = "WORKFLOW_EXECUTION_STARTED"
    WORKFLOW_EXECUTION_COMPLETED = "WORKFLOW_EXECUTION_COMPLETED"
    WORKFLOW_EXECUTION_FAILED = "WORKFLOW_EXECUTION_FAILED"
    WORKFLOW_EXECUTION_CANCELED = "WORKFLOW_EXECUTION_CANCELED"
    WORKFLOW_EXECUTION_CONTINUED_AS_NEW = "WORKFLOW_EXECUTION_CONTINUED_AS_NEW"
    WORKFLOW_TASK_TIMED_OUT = "WORKFLOW_TASK_TIMED_OUT"
    WORKFLOW_TASK_FAILED = "WORKFLOW_TASK_FAILED"
    CUSTOM_TASK_STARTED = "CUSTOM_TASK_STARTED"
    CUSTOM_TASK_IN_PROGRESS = "CUSTOM_TASK_IN_PROGRESS"
    CUSTOM_TASK_COMPLETED = "CUSTOM_TASK_COMPLETED"
    CUSTOM_TASK_FAILED = "CUSTOM_TASK_FAILED"
    CUSTOM_TASK_TIMED_OUT = "CUSTOM_TASK_TIMED_OUT"
    CUSTOM_TASK_CANCELED = "CUSTOM_TASK_CANCELED"
    ACTIVITY_TASK_STARTED = "ACTIVITY_TASK_STARTED"
    ACTIVITY_TASK_COMPLETED = "ACTIVITY_TASK_COMPLETED"
    ACTIVITY_TASK_RETRYING = "ACTIVITY_TASK_RETRYING"
    ACTIVITY_TASK_FAILED = "ACTIVITY_TASK_FAILED"


class JSONPatchBase(BaseModel):
    """
    A single JSON Patch operation as defined in RFC 6902, with extensions.

    Used for incremental updates to JSON documents without replacing the entire payload.

    Extensions beyond RFC 6902:
    - "append" operation: Optimized string concatenation for streaming text updates.
    - path can be an empty list [] for root-level operations (mutative compatibility).
    """

    path: str = Field(
        description="A JSON Pointer (RFC 6901) identifying the target location within the document. "
        "Can be a string path (e.g., '/foo/bar'), '/', '', or an empty list [] for root-level operations."
    )
    value: Any = Field(description="The value to use for the operation")


class JSONPatchAdd(JSONPatchBase):
    op: Literal["add"] = Field(description="Add operation ")


class JSONPatchReplace(JSONPatchBase):
    op: Literal["replace"] = Field(description="Replace operation")


class JSONPatchRemove(JSONPatchBase):
    op: Literal["remove"] = Field(description="Add operation ")


class JSONPatchAppend(JSONPatchBase):
    op: Literal["append"] = Field(
        description="'append' is an extension for efficient string concatenation in streaming scenarios."
    )
    value: str = Field(description="The value to use for the operation. A string to append to the existing value")


JSONPatch = Annotated[
    Annotated[JSONPatchAppend, Tag("append")]
    | Annotated[JSONPatchAdd, Tag("add")]
    | Annotated[JSONPatchReplace, Tag("replace")]
    | Annotated[JSONPatchRemove, Tag("remove")],
    Discriminator("op"),
]


def json_patch(op: str, path: str, value: Any) -> JSONPatch:
    if op == "add":
        return JSONPatchAdd(op="add", path=path, value=value)
    elif op == "replace":
        return JSONPatchReplace(op="replace", path=path, value=value)
    elif op == "remove":
        return JSONPatchRemove(op="remove", path=path, value=value)
    elif op == "append":
        return JSONPatchAppend(op="append", path=path, value=value)
    else:
        raise ValueError(f"Unknown operation: {op}")


class JSONPatchPayload(BaseModel):
    """
    A payload containing a list of JSON Patch operations.

    Used for streaming incremental updates to workflow state.
    """

    type: Literal["json_patch"] = Field(
        default="json_patch",
        description="Discriminator indicating this is a JSON Patch payload.",
    )
    value: list[JSONPatch] = Field(description="The list of JSON Patch operations to apply in order.")

    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class JSONPayload(BaseModel):
    """
    A payload containing arbitrary JSON data.

    Used for complete state snapshots or final results.
    """

    type: Literal["json"] = Field(
        default="json",
        description="Discriminator indicating this is a raw JSON payload.",
    )
    value: Any = Field(description="The JSON-serializable payload value.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


Payload = Annotated[
    Annotated[JSONPayload, Tag("json")] | Annotated[JSONPatchPayload, Tag("json_patch")],
    Discriminator("type"),
]
"""Union type for payload data, discriminated by the 'type' field."""


class Failure(BaseModel):
    """Represents an error or exception that occurred during execution."""

    message: str = Field(description="A human-readable description of the failure.")


def _get_timestamp_ns() -> int:
    """Get current timestamp in nanoseconds, workflow-safe."""
    if temporalio.workflow.in_workflow():
        return temporalio.workflow.time_ns()
    return time.time_ns()


class BaseEvent(BaseModel):
    """
    Base class for all workflow events.

    All events share common identification and timing fields to enable
    correlation and ordering across distributed workflow executions.
    """

    event_id: str = Field(description="Unique identifier for this event instance.")
    event_timestamp: int = Field(
        default_factory=_get_timestamp_ns,
        description="Unix timestamp in nanoseconds when the event was created.",
    )
    root_workflow_exec_id: str = Field(
        description="Execution ID of the root workflow that initiated this execution chain."
    )
    parent_workflow_exec_id: str | None = Field(
        default=None,
        description="Execution ID of the parent workflow that initiated this execution. "
        "If this is a root workflow, this field is not set.",
    )
    workflow_exec_id: str = Field(description="Execution ID of the workflow that emitted this event.")
    workflow_run_id: str = Field(
        description="Run ID of the workflow execution. "
        "Changes on continue-as-new while workflow_exec_id stays the same."
    )
    workflow_name: str = Field(description="The registered name of the workflow that emitted this event.")

    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class BaseTaskAttributes(BaseModel):
    """Base attributes shared by all workflow task events."""

    task_id: str = Field(description="Unique identifier for the task within the workflow execution.")


class WorkflowExecutionStartedAttributes(BaseTaskAttributes):
    """Attributes for workflow execution started events."""

    workflow_name: str = Field(description="The registered name of the workflow being executed.")
    input: JSONPayload = Field(description="The input arguments passed to the workflow.")


class WorkflowExecutionCompletedAttributes(BaseTaskAttributes):
    """Attributes for workflow execution completed events."""

    result: JSONPayload = Field(description="The final result returned by the workflow.")


class WorkflowExecutionFailedAttributes(BaseTaskAttributes):
    """Attributes for workflow execution failed events."""

    failure: Failure = Field(description="Details about the failure that caused the workflow to fail.")


class WorkflowExecutionCanceledAttributes(BaseTaskAttributes):
    """Attributes for workflow execution canceled events."""

    reason: str | None = Field(
        default=None,
        description="Optional reason provided for the cancellation.",
    )


class WorkflowExecutionContinuedAsNewAttributes(BaseTaskAttributes):
    """Attributes for workflow execution continued-as-new events."""

    new_execution_run_id: str = Field(
        description="The run ID of the new workflow execution that continues this workflow."
    )
    workflow_name: str = Field(description="The registered name of the continued workflow.")
    input: JSONPayload = Field(description="The input arguments passed to the new workflow execution.")


class WorkflowTaskTimedOutAttributes(BaseTaskAttributes):
    """Attributes for workflow task timed out events."""

    timeout_type: str | None = Field(
        default=None,
        description="The type of timeout that occurred (e.g., 'START_TO_CLOSE', 'SCHEDULE_TO_START').",
    )


class WorkflowTaskFailedAttributes(BaseTaskAttributes):
    """Attributes for workflow task failed events."""

    failure: Failure = Field(description="Details about the failure that caused the task to fail.")


class BaseCustomTaskAttributes(BaseModel):
    """Base attributes shared by all custom task events."""

    custom_task_id: str = Field(description="Unique identifier for the custom task within the workflow.")
    custom_task_type: str = Field(description="The type/category of the custom task (e.g., 'llm_call', 'api_request').")


class CustomTaskStartedAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task started events."""

    payload: JSONPayload = Field(
        default_factory=lambda: JSONPayload(value=None), description="The initial state or payload for the custom task."
    )


class CustomTaskInProgressAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task in-progress events with streaming updates."""

    payload: Payload = Field(description="The current state or incremental update for the task.")


class CustomTaskCompletedAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task completed events."""

    payload: JSONPayload = Field(description="The final result of the custom task.")


class CustomTaskFailedAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task failed events."""

    failure: Failure = Field(description="Details about the failure that caused the task to fail.")


class CustomTaskTimedOutAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task timed out events."""

    timeout_type: str | None = Field(
        default=None,
        description="The type of timeout that occurred.",
    )


class CustomTaskCanceledAttributes(BaseCustomTaskAttributes):
    """Attributes for custom task canceled events."""

    reason: str | None = Field(
        default=None,
        description="Optional reason provided for the cancellation.",
    )


class BaseActivityTaskAttributes(BaseModel):
    """Base attributes shared by all activity task events."""

    task_id: str = Field(description="Unique identifier for the activity task within the workflow.")
    activity_name: str = Field(description="The registered name of the activity being executed.")


class ActivityTaskStartedAttributes(BaseActivityTaskAttributes):
    """Attributes for activity task started events."""

    input: JSONPayload = Field(description="The input arguments passed to the activity.")


class ActivityTaskCompletedAttributes(BaseActivityTaskAttributes):
    """Attributes for activity task completed events."""

    result: JSONPayload = Field(description="The result returned by the activity.")


class ActivityTaskRetryingAttributes(BaseActivityTaskAttributes):
    """Attributes for activity task retrying events."""

    attempt: int = Field(description="The attempt number that failed (1-indexed).")
    failure: Failure = Field(description="Details about the failure that caused the retry.")


class ActivityTaskFailedAttributes(BaseActivityTaskAttributes):
    """Attributes for activity task failed events (final failure after all retries)."""

    attempt: int = Field(description="The final attempt number that failed (1-indexed).")
    failure: Failure = Field(description="Details about the failure that caused the activity to fail.")


class WorkflowExecutionStarted(BaseEvent):
    """
    Emitted when a workflow execution begins.

    This is the first event in any workflow execution lifecycle.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_EXECUTION_STARTED] = Field(
        default=WorkflowEventType.WORKFLOW_EXECUTION_STARTED,
        description="Event type discriminator.",
    )
    attributes: WorkflowExecutionStartedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowExecutionCompleted(BaseEvent):
    """
    Emitted when a workflow execution completes successfully.

    This is a terminal event indicating the workflow finished without errors.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED] = Field(
        default=WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED,
        description="Event type discriminator.",
    )
    attributes: WorkflowExecutionCompletedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowExecutionFailed(BaseEvent):
    """
    Emitted when a workflow execution fails due to an unhandled exception.

    This is a terminal event indicating the workflow ended with an error.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_EXECUTION_FAILED] = Field(
        default=WorkflowEventType.WORKFLOW_EXECUTION_FAILED,
        description="Event type discriminator.",
    )
    attributes: WorkflowExecutionFailedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowExecutionCanceled(BaseEvent):
    """
    Emitted when a workflow execution is canceled.

    This is a terminal event indicating the workflow was explicitly canceled.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_EXECUTION_CANCELED] = Field(
        default=WorkflowEventType.WORKFLOW_EXECUTION_CANCELED,
        description="Event type discriminator.",
    )
    attributes: WorkflowExecutionCanceledAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowExecutionContinuedAsNew(BaseEvent):
    """
    Emitted when a workflow continues as a new execution.

    This occurs when a workflow uses continue-as-new to reset its history
    while maintaining logical continuity.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_EXECUTION_CONTINUED_AS_NEW] = Field(
        default=WorkflowEventType.WORKFLOW_EXECUTION_CONTINUED_AS_NEW,
        description="Event type discriminator.",
    )
    attributes: WorkflowExecutionContinuedAsNewAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowTaskTimedOut(BaseEvent):
    """
    Emitted when a workflow task times out.

    This indicates the workflow task (a unit of workflow execution) exceeded
    its configured timeout.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_TASK_TIMED_OUT] = Field(
        default=WorkflowEventType.WORKFLOW_TASK_TIMED_OUT,
        description="Event type discriminator.",
    )
    attributes: WorkflowTaskTimedOutAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class WorkflowTaskFailed(BaseEvent):
    """
    Emitted when a workflow task fails.

    This indicates an error occurred during workflow task execution,
    which may trigger a retry depending on configuration.
    """

    event_type: Literal[WorkflowEventType.WORKFLOW_TASK_FAILED] = Field(
        default=WorkflowEventType.WORKFLOW_TASK_FAILED,
        description="Event type discriminator.",
    )
    attributes: WorkflowTaskFailedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskStarted(BaseEvent):
    """
    Emitted when a custom task begins execution.

    Custom tasks represent user-defined units of work within a workflow,
    such as LLM calls, API requests, or data processing steps.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_STARTED] = Field(
        default=WorkflowEventType.CUSTOM_TASK_STARTED,
        description="Event type discriminator.",
    )
    attributes: CustomTaskStartedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskInProgress(BaseEvent):
    """
    Emitted during custom task execution to report progress.

    This event supports streaming updates via JSON or JSON Patch payloads,
    enabling real-time progress tracking for long-running tasks.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_IN_PROGRESS] = Field(
        default=WorkflowEventType.CUSTOM_TASK_IN_PROGRESS,
        description="Event type discriminator.",
    )
    attributes: CustomTaskInProgressAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskCompleted(BaseEvent):
    """
    Emitted when a custom task completes successfully.

    Contains the final result of the task execution.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_COMPLETED] = Field(
        default=WorkflowEventType.CUSTOM_TASK_COMPLETED,
        description="Event type discriminator.",
    )
    attributes: CustomTaskCompletedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskFailed(BaseEvent):
    """
    Emitted when a custom task fails.

    Contains details about the failure for debugging and error handling.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_FAILED] = Field(
        default=WorkflowEventType.CUSTOM_TASK_FAILED,
        description="Event type discriminator.",
    )
    attributes: CustomTaskFailedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskTimedOut(BaseEvent):
    """
    Emitted when a custom task exceeds its timeout.

    Indicates the task did not complete within its configured time limit.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_TIMED_OUT] = Field(
        default=WorkflowEventType.CUSTOM_TASK_TIMED_OUT,
        description="Event type discriminator.",
    )
    attributes: CustomTaskTimedOutAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class CustomTaskCanceled(BaseEvent):
    """
    Emitted when a custom task is canceled.

    Indicates the task was explicitly stopped before completion.
    """

    event_type: Literal[WorkflowEventType.CUSTOM_TASK_CANCELED] = Field(
        default=WorkflowEventType.CUSTOM_TASK_CANCELED,
        description="Event type discriminator.",
    )
    attributes: CustomTaskCanceledAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class ActivityTaskStarted(BaseEvent):
    """
    Emitted when an activity task begins execution.

    This is the first event for an activity, emitted on the first attempt only.
    Subsequent retry attempts emit ACTIVITY_TASK_RETRYING instead.
    """

    event_type: Literal[WorkflowEventType.ACTIVITY_TASK_STARTED] = Field(
        default=WorkflowEventType.ACTIVITY_TASK_STARTED,
        description="Event type discriminator.",
    )
    attributes: ActivityTaskStartedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class ActivityTaskCompleted(BaseEvent):
    """
    Emitted when an activity task completes successfully.

    Contains timing information about the successful execution.
    """

    event_type: Literal[WorkflowEventType.ACTIVITY_TASK_COMPLETED] = Field(
        default=WorkflowEventType.ACTIVITY_TASK_COMPLETED,
        description="Event type discriminator.",
    )
    attributes: ActivityTaskCompletedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class ActivityTaskRetrying(BaseEvent):
    """
    Emitted when an activity task fails and will be retried.

    Contains information about the failed attempt and the error that occurred.
    """

    event_type: Literal[WorkflowEventType.ACTIVITY_TASK_RETRYING] = Field(
        default=WorkflowEventType.ACTIVITY_TASK_RETRYING,
        description="Event type discriminator.",
    )
    attributes: ActivityTaskRetryingAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


class ActivityTaskFailed(BaseEvent):
    """
    Emitted when an activity task fails after exhausting all retry attempts.

    This is a terminal event indicating the activity could not complete successfully.
    """

    event_type: Literal[WorkflowEventType.ACTIVITY_TASK_FAILED] = Field(
        default=WorkflowEventType.ACTIVITY_TASK_FAILED,
        description="Event type discriminator.",
    )
    attributes: ActivityTaskFailedAttributes = Field(description="Event-specific attributes.")
    model_config = ConfigDict(json_schema_serialization_defaults_required=True)


def _get_event_type_discriminator(v: Any) -> str:
    """Extract event type for discriminated union parsing."""
    if isinstance(v, dict):
        event_type_val = v.get("event_type", "")
        # Handle both string and WorkflowEventType enum values
        if isinstance(event_type_val, WorkflowEventType):
            return event_type_val.value
        return str(event_type_val)

    event_type_attr = getattr(v, "event_type", "")
    if isinstance(event_type_attr, WorkflowEventType):
        return event_type_attr.value
    return str(event_type_attr)


WorkflowExecutionEvent = Annotated[
    Annotated[WorkflowExecutionStarted, Tag(WorkflowEventType.WORKFLOW_EXECUTION_STARTED.value)]
    | Annotated[WorkflowExecutionCompleted, Tag(WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED.value)]
    | Annotated[WorkflowExecutionFailed, Tag(WorkflowEventType.WORKFLOW_EXECUTION_FAILED.value)]
    | Annotated[WorkflowExecutionCanceled, Tag(WorkflowEventType.WORKFLOW_EXECUTION_CANCELED.value)]
    | Annotated[WorkflowExecutionContinuedAsNew, Tag(WorkflowEventType.WORKFLOW_EXECUTION_CONTINUED_AS_NEW.value)]
    | Annotated[WorkflowTaskTimedOut, Tag(WorkflowEventType.WORKFLOW_TASK_TIMED_OUT.value)]
    | Annotated[WorkflowTaskFailed, Tag(WorkflowEventType.WORKFLOW_TASK_FAILED.value)],
    Discriminator(_get_event_type_discriminator),
]
"""
Union of all workflow execution events, discriminated by 'event_type'.

Use this type when parsing workflow lifecycle events.
"""


CustomTaskEvent = Annotated[
    Annotated[CustomTaskStarted, Tag(WorkflowEventType.CUSTOM_TASK_STARTED.value)]
    | Annotated[CustomTaskInProgress, Tag(WorkflowEventType.CUSTOM_TASK_IN_PROGRESS.value)]
    | Annotated[CustomTaskCompleted, Tag(WorkflowEventType.CUSTOM_TASK_COMPLETED.value)]
    | Annotated[CustomTaskFailed, Tag(WorkflowEventType.CUSTOM_TASK_FAILED.value)]
    | Annotated[CustomTaskTimedOut, Tag(WorkflowEventType.CUSTOM_TASK_TIMED_OUT.value)]
    | Annotated[CustomTaskCanceled, Tag(WorkflowEventType.CUSTOM_TASK_CANCELED.value)],
    Discriminator(_get_event_type_discriminator),
]
"""
Union of all custom task events, discriminated by 'event_type'.

Use this type when parsing custom task lifecycle events.
"""


ActivityTaskEvent = Annotated[
    Annotated[ActivityTaskStarted, Tag(WorkflowEventType.ACTIVITY_TASK_STARTED.value)]
    | Annotated[ActivityTaskCompleted, Tag(WorkflowEventType.ACTIVITY_TASK_COMPLETED.value)]
    | Annotated[ActivityTaskRetrying, Tag(WorkflowEventType.ACTIVITY_TASK_RETRYING.value)]
    | Annotated[ActivityTaskFailed, Tag(WorkflowEventType.ACTIVITY_TASK_FAILED.value)],
    Discriminator(_get_event_type_discriminator),
]
"""
Union of all activity task events, discriminated by 'event_type'.

Use this type when parsing activity task lifecycle events.
"""


WorkflowEvent = Annotated[
    Annotated[WorkflowExecutionStarted, Tag(WorkflowEventType.WORKFLOW_EXECUTION_STARTED.value)]
    | Annotated[WorkflowExecutionCompleted, Tag(WorkflowEventType.WORKFLOW_EXECUTION_COMPLETED.value)]
    | Annotated[WorkflowExecutionFailed, Tag(WorkflowEventType.WORKFLOW_EXECUTION_FAILED.value)]
    | Annotated[WorkflowExecutionCanceled, Tag(WorkflowEventType.WORKFLOW_EXECUTION_CANCELED.value)]
    | Annotated[WorkflowExecutionContinuedAsNew, Tag(WorkflowEventType.WORKFLOW_EXECUTION_CONTINUED_AS_NEW.value)]
    | Annotated[WorkflowTaskTimedOut, Tag(WorkflowEventType.WORKFLOW_TASK_TIMED_OUT.value)]
    | Annotated[WorkflowTaskFailed, Tag(WorkflowEventType.WORKFLOW_TASK_FAILED.value)]
    | Annotated[CustomTaskStarted, Tag(WorkflowEventType.CUSTOM_TASK_STARTED.value)]
    | Annotated[CustomTaskInProgress, Tag(WorkflowEventType.CUSTOM_TASK_IN_PROGRESS.value)]
    | Annotated[CustomTaskCompleted, Tag(WorkflowEventType.CUSTOM_TASK_COMPLETED.value)]
    | Annotated[CustomTaskFailed, Tag(WorkflowEventType.CUSTOM_TASK_FAILED.value)]
    | Annotated[CustomTaskTimedOut, Tag(WorkflowEventType.CUSTOM_TASK_TIMED_OUT.value)]
    | Annotated[CustomTaskCanceled, Tag(WorkflowEventType.CUSTOM_TASK_CANCELED.value)]
    | Annotated[ActivityTaskStarted, Tag(WorkflowEventType.ACTIVITY_TASK_STARTED.value)]
    | Annotated[ActivityTaskCompleted, Tag(WorkflowEventType.ACTIVITY_TASK_COMPLETED.value)]
    | Annotated[ActivityTaskRetrying, Tag(WorkflowEventType.ACTIVITY_TASK_RETRYING.value)]
    | Annotated[ActivityTaskFailed, Tag(WorkflowEventType.ACTIVITY_TASK_FAILED.value)],
    Discriminator(_get_event_type_discriminator),
]
"""
Union of all workflow events (execution + custom task + activity task), discriminated by 'event_type'.

This is the primary type for parsing any event from the workflow event stream.

Example:
    ```python
    from pydantic import TypeAdapter

    event_adapter = TypeAdapter(WorkflowEvent)
    event = event_adapter.validate_python(event_dict)
    ```
"""


class WorkflowEventRequest(BaseModel):
    """Request model containing a workflow event."""

    event: WorkflowEvent = Field(description="The workflow event payload.")


class ListWorkflowEventResponse(BaseModel):
    events: list[WorkflowEvent] = Field(description="List of workflow events.")
    next_cursor: str | None = Field(default=None, description="Cursor for pagination.")


class WorkflowEventResponse(BaseModel):
    """Response model for workflow event reception."""

    status: Literal["success", "error"] = Field(description="Status of the event reception")
    message: str | None = Field(default=None, description="Optional message")


class EventSource(StrEnum):
    DATABASE = "DATABASE"
    LIVE = "LIVE"


__all__ = [
    # Enums
    "WorkflowEventType",
    # Payload models
    "JSONPatch",
    "JSONPatchPayload",
    "JSONPayload",
    "Payload",
    # Failure model
    "Failure",
    # Base models
    "BaseEvent",
    "BaseTaskAttributes",
    "BaseCustomTaskAttributes",
    "BaseActivityTaskAttributes",
    # Workflow execution attributes
    "WorkflowExecutionStartedAttributes",
    "WorkflowExecutionCompletedAttributes",
    "WorkflowExecutionFailedAttributes",
    "WorkflowExecutionCanceledAttributes",
    "WorkflowExecutionContinuedAsNewAttributes",
    "WorkflowTaskTimedOutAttributes",
    "WorkflowTaskFailedAttributes",
    # Custom task attributes
    "CustomTaskStartedAttributes",
    "CustomTaskInProgressAttributes",
    "CustomTaskCompletedAttributes",
    "CustomTaskFailedAttributes",
    "CustomTaskTimedOutAttributes",
    "CustomTaskCanceledAttributes",
    # Activity task attributes
    "ActivityTaskStartedAttributes",
    "ActivityTaskCompletedAttributes",
    "ActivityTaskRetryingAttributes",
    "ActivityTaskFailedAttributes",
    # Workflow execution events
    "WorkflowExecutionStarted",
    "WorkflowExecutionCompleted",
    "WorkflowExecutionFailed",
    "WorkflowExecutionCanceled",
    "WorkflowExecutionContinuedAsNew",
    "WorkflowTaskTimedOut",
    "WorkflowTaskFailed",
    # Custom task events
    "CustomTaskStarted",
    "CustomTaskInProgress",
    "CustomTaskCompleted",
    "CustomTaskFailed",
    "CustomTaskTimedOut",
    "CustomTaskCanceled",
    # Activity task events
    "ActivityTaskStarted",
    "ActivityTaskCompleted",
    "ActivityTaskRetrying",
    "ActivityTaskFailed",
    # Discriminated unions
    "WorkflowExecutionEvent",
    "CustomTaskEvent",
    "ActivityTaskEvent",
    "WorkflowEvent",
    # Request/Response models
    "WorkflowEventRequest",
    "WorkflowEventResponse",
]
